@@ -25,10 +25,10 @@
 #include "nsXPCOMCIDInternal.h"
 
 #include "mozilla/LookAndFeel.h"
-#include "mozilla/Unused.h"
 
-#include "nsIAppStartup.h"
 #include "nsIObserverService.h"
+#include "mozilla/AppShutdown.h"
+#include "mozilla/Components.h"
 #include "mozilla/Preferences.h"
 #include "nsIResProtocolHandler.h"
 #include "nsIScriptError.h"
@@ -202,8 +202,7 @@ nsChromeRegistryChrome::Observe(nsISupports* aSubject, const char* aTopic,
 
 NS_IMETHODIMP
 nsChromeRegistryChrome::CheckForNewChrome() {
-  nsCOMPtr<nsIAppStartup> appStartup = components::AppStartup::Service();
-  if (appStartup->GetShuttingDown()) {
+  if (AppShutdown::IsInOrBeyond(ShutdownPhase::AppShutdownConfirmed)) {
     MOZ_ASSERT(false, "checking for new chrome during shutdown");
     return NS_ERROR_UNEXPECTED;
   }
@@ -233,10 +232,10 @@ void nsChromeRegistryChrome::SendRegisteredChrome(
   nsTArray<SubstitutionMapping> resources;
   nsTArray<OverrideMapping> overrides;
 
-  for (auto iter = mPackagesHash.Iter(); !iter.Done(); iter.Next()) {
+  for (const auto& entry : mPackagesHash) {
     ChromePackage chromePackage;
-    ChromePackageFromPackageEntry(iter.Key(), iter.UserData(), &chromePackage,
-                                  SKIN);
+    ChromePackageFromPackageEntry(entry.GetKey(), entry.GetWeak(),
+                                  &chromePackage, SKIN);
     packages.AppendElement(chromePackage);
   }
 
@@ -257,14 +256,14 @@ void nsChromeRegistryChrome::SendRegisteredChrome(
     NS_ENSURE_SUCCESS_VOID(rv);
   }
 
-  for (auto iter = mOverrideTable.Iter(); !iter.Done(); iter.Next()) {
+  for (const auto& entry : mOverrideTable) {
     SerializedURI chromeURI, overrideURI;
 
-    SerializeURI(iter.Key(), chromeURI);
-    SerializeURI(iter.UserData(), overrideURI);
+    SerializeURI(entry.GetKey(), chromeURI);
+    SerializeURI(entry.GetWeak(), overrideURI);
 
-    OverrideMapping override = {chromeURI, overrideURI};
-    overrides.AppendElement(override);
+    overrides.AppendElement(
+        OverrideMapping{std::move(chromeURI), std::move(overrideURI)});
   }
 
   nsAutoCString appLocale;
@@ -329,9 +328,13 @@ nsIURI* nsChromeRegistryChrome::GetBaseURIFromPackage(
     nsAutoCString appLocale;
     LocaleService::GetInstance()->GetAppLocaleAsLangTag(appLocale);
     return entry->locales.GetBase(appLocale, nsProviderArray::LOCALE);
-  } else if (aProvider.EqualsLiteral("skin")) {
+  }
+
+  if (aProvider.EqualsLiteral("skin")) {
     return entry->skins.GetBase(SKIN, nsProviderArray::ANY);
-  } else if (aProvider.EqualsLiteral("content")) {
+  }
+
+  if (aProvider.EqualsLiteral("content")) {
     return entry->baseURI;
   }
   return nullptr;
@@ -447,7 +450,7 @@ static void SendManifestEntry(const ChromeRegistryItem& aItem) {
   if (!parents.Length()) return;
 
   for (uint32_t i = 0; i < parents.Length(); i++) {
-    Unused << parents[i]->SendRegisterChromeItem(aItem);
+    (void)parents[i]->SendRegisterChromeItem(aItem);
   }
 }
 
@@ -476,7 +479,7 @@ void nsChromeRegistryChrome::ManifestContent(ManifestProcessingContext& cx,
   }
 
   nsDependentCString packageName(package);
-  PackageEntry* entry = mPackagesHash.LookupOrAdd(packageName);
+  PackageEntry* entry = mPackagesHash.GetOrInsertNew(packageName);
   entry->baseURI = resolved;
   entry->flags = flags;
 
@@ -513,7 +516,7 @@ void nsChromeRegistryChrome::ManifestLocale(ManifestProcessingContext& cx,
   }
 
   nsDependentCString packageName(package);
-  PackageEntry* entry = mPackagesHash.LookupOrAdd(packageName);
+  PackageEntry* entry = mPackagesHash.GetOrInsertNew(packageName);
   entry->locales.SetBase(nsDependentCString(provider), resolved);
 
   if (mDynamicRegistration) {
@@ -558,7 +561,7 @@ void nsChromeRegistryChrome::ManifestSkin(ManifestProcessingContext& cx,
   }
 
   nsDependentCString packageName(package);
-  PackageEntry* entry = mPackagesHash.LookupOrAdd(packageName);
+  PackageEntry* entry = mPackagesHash.GetOrInsertNew(packageName);
   entry->skins.SetBase(nsDependentCString(provider), resolved);
 
   if (mDynamicRegistration) {
@@ -583,33 +586,13 @@ void nsChromeRegistryChrome::ManifestOverride(ManifestProcessingContext& cx,
     return;
   }
 
-  if (cx.mType == NS_SKIN_LOCATION) {
-    bool chromeSkinOnly =
-        chromeuri->SchemeIs("chrome") && resolveduri->SchemeIs("chrome");
-    if (chromeSkinOnly) {
-      nsAutoCString chromePath, resolvedPath;
-      chromeuri->GetPathQueryRef(chromePath);
-      resolveduri->GetPathQueryRef(resolvedPath);
-      chromeSkinOnly = StringBeginsWith(chromePath, "/skin/"_ns) &&
-                       StringBeginsWith(resolvedPath, "/skin/"_ns);
-    }
-    if (!chromeSkinOnly) {
-      LogMessageWithContext(
-          cx.GetManifestURI(), lineno, nsIScriptError::warningFlag,
-          "Cannot register non-chrome://.../skin/ URIs '%s' and '%s' as "
-          "overrides and/or to be overridden from a skin manifest.",
-          chrome, resolved);
-      return;
-    }
-  }
-
   if (!CanLoadResource(resolveduri)) {
     LogMessageWithContext(
         cx.GetManifestURI(), lineno, nsIScriptError::warningFlag,
         "Cannot register non-local URI '%s' for an override.", resolved);
     return;
   }
-  mOverrideTable.Put(chromeuri, resolveduri);
+  mOverrideTable.InsertOrUpdate(chromeuri, resolveduri);
 
   if (mDynamicRegistration) {
     SerializedURI serializedChrome;
@@ -632,7 +615,7 @@ void nsChromeRegistryChrome::ManifestResource(ManifestProcessingContext& cx,
   EnsureLowerCase(package);
   nsDependentCString host(package);
 
-  nsCOMPtr<nsIIOService> io = mozilla::services::GetIOService();
+  nsCOMPtr<nsIIOService> io = mozilla::components::IO::Service();
   if (!io) {
     NS_WARNING("No IO service trying to process chrome manifests");
     return;
