@@ -19,7 +19,6 @@
 #include "nsIFrameInlines.h"
 #include "nsPlaceholderFrame.h"
 #include "nsSubDocumentFrame.h"
-#include "nsViewManager.h"
 
 /**
  * Code for doing display list building for a modified subset of the window,
@@ -384,18 +383,19 @@ static Maybe<const ActiveScrolledRoot*> SelectContainerASR(
 
 static void UpdateASR(nsDisplayItem* aItem,
                       Maybe<const ActiveScrolledRoot*>& aContainerASR) {
+  const Maybe<const ActiveScrolledRoot*> frameASR =
+      aItem->GetBaseASRForAncestorOfContainedASR();
+  if (!frameASR) {
+    return;
+  }
+
   if (!aContainerASR) {
+    aItem->SetActiveScrolledRoot(*frameASR);
     return;
   }
 
-  nsDisplayWrapList* wrapList = aItem->AsDisplayWrapList();
-  if (!wrapList) {
-    aItem->SetActiveScrolledRoot(*aContainerASR);
-    return;
-  }
-
-  wrapList->SetActiveScrolledRoot(ActiveScrolledRoot::PickAncestor(
-      wrapList->GetFrameActiveScrolledRoot(), *aContainerASR));
+  aItem->SetActiveScrolledRoot(
+      ActiveScrolledRoot::PickAncestor(*frameASR, *aContainerASR));
 }
 
 static void CopyASR(nsDisplayItem* aOld, nsDisplayItem* aNew) {
