@@ -7,9 +7,6 @@
 #ifndef nsPrintData_h___
 #define nsPrintData_h___
 
-#include "mozilla/Attributes.h"
-#include "mozilla/UniquePtr.h"
-
 // Interfaces
 #include "nsCOMArray.h"
 #include "nsDeviceContext.h"
