@@ -435,6 +435,13 @@ bool ComputedStyle::HasAnchorPosReference() const {
   if (pos->mPositionAnchor.IsIdent()) {
     // Short circuit if there's a default anchor defined, even if
     // it may not end up being referenced.
+    // If this early return is removed, we'll need to handle mPositionArea
+    // explicitly.
+    return true;
+  }
+
+  if (!pos->mPositionArea.IsNone()) {
+    // Position area is relative to an anchor.
     return true;
   }
 
@@ -446,13 +453,43 @@ bool ComputedStyle::HasAnchorPosReference() const {
     return aInset.HasAnchorPositioningFunction();
   }) || pos->mWidth.HasAnchorPositioningFunction() ||
          pos->mHeight.HasAnchorPositioningFunction() ||
+         pos->mMinWidth.HasAnchorPositioningFunction() ||
          pos->mMinHeight.HasAnchorPositioningFunction() ||
-         pos->mMinHeight.HasAnchorPositioningFunction() ||
-         pos->mMaxHeight.HasAnchorPositioningFunction() ||
+         pos->mMaxWidth.HasAnchorPositioningFunction() ||
          pos->mMaxHeight.HasAnchorPositioningFunction() ||
          StyleMargin()->mMargin.Any([](const ::mozilla::StyleMargin& aMargin) {
            return aMargin.HasAnchorPositioningFunction();
          });
 }
 
+// XXX: This is a broad-stroke method to return true if the referenced anchors
+// in two computed style may differ. Since we do not get the actual anchor
+// names, we do not know if the difference is just a position/size/margin change
+// or if indeed the anchors changed. So we will get false positives here, hence
+// "maybe".
+bool ComputedStyle::MaybeAnchorPosReferencesDiffer(
+    const ComputedStyle* aOther) const {
+  if (!HasAnchorPosReference() || !aOther->HasAnchorPosReference()) {
+    return true;
+  }
+
+  const auto* pos = StylePosition();
+  const auto* otherPos = aOther->StylePosition();
+  if (pos->mOffset != otherPos->mOffset || pos->mWidth != otherPos->mWidth ||
+      pos->mHeight != otherPos->mHeight ||
+      pos->mMinWidth != otherPos->mMinWidth ||
+      pos->mMinHeight != otherPos->mMinHeight ||
+      pos->mMaxWidth != otherPos->mMaxWidth ||
+      pos->mMaxHeight != otherPos->mMaxHeight ||
+      pos->mPositionAnchor != otherPos->mPositionAnchor) {
+    return true;
+  }
+
+  if (StyleMargin()->mMargin != aOther->StyleMargin()->mMargin) {
+    return true;
+  }
+
+  return false;
+}
+
 }  // namespace mozilla
