@@ -9,6 +9,7 @@
 #include <algorithm>
 
 #include "LayoutConstants.h"
+#include "WindowRenderer.h"
 #include "X11UndefineNone.h"
 #include "XULButtonElement.h"
 #include "XULPopupElement.h"
@@ -110,7 +111,7 @@ class PopupExpirationTracker final
 
  public:
   PopupExpirationTracker()
-      : nsExpirationTracker(5000 /* ms */, "PopupExpirationTracker") {}
+      : nsExpirationTracker(5000 /* ms */, "PopupExpirationTracker"_ns) {}
   static PopupExpirationTracker* Get() { return sInstance.get(); }
   static PopupExpirationTracker& GetOrCreate() {
     if (!sInstance) {
@@ -171,10 +172,6 @@ void nsMenuPopupFrame::Init(nsIContent* aContent, nsContainerFrame* aParent,
                             nsIFrame* aPrevInFlow) {
   nsBlockFrame::Init(aContent, aParent, aPrevInFlow);
 
-  CreatePopupView();
-
-  nsView* ourView = GetView();
-
   const auto& el = PopupElement();
   mPopupType = PopupType::Panel;
   if (el.IsMenu()) {
@@ -198,8 +195,10 @@ void nsMenuPopupFrame::Init(nsIContent* aContent, nsContainerFrame* aParent,
     }
   }
 
-  if (!ourView->HasWidget() && ShouldHaveWidgetWhenHidden()) {
-    CreateWidgetForView(ourView);
+  // To improve performance, create the widget for the popup if needed. Popups
+  // such as menus will create their widgets later when the popup opens.
+  if (!mWidget && ShouldHaveWidgetWhenHidden()) {
+    CreateWidget();
   }
 
   AddStateBits(NS_FRAME_IN_POPUP);
@@ -262,19 +261,15 @@ void nsMenuPopupFrame::PrepareWidget(bool aForceRecreate) {
   if (mExpirationState.IsTracked()) {
     PopupExpirationTracker::Get()->RemoveObject(this);
   }
-  nsView* ourView = GetView();
   if (auto* widget = GetWidget()) {
     nsCOMPtr<nsIWidget> parent = ComputeParentWidget();
     if (aForceRecreate || widget->GetParent() != parent ||
         widget->NeedsRecreateToReshow()) {
-      // Widget's WebRender resources needs to be cleared before creating new
-      // widget.
-      widget->ClearCachedWebrenderResources();
-      ourView->DestroyWidget();
+      DestroyWidget();
     }
   }
-  if (!ourView->HasWidget()) {
-    CreateWidgetForView(ourView);
+  if (!mWidget) {
+    CreateWidget();
   } else {
     PropagateStyleToWidget();
   }
@@ -302,13 +297,13 @@ already_AddRefed<nsIWidget> nsMenuPopupFrame::ComputeParentWidget() const {
       parentWidget = baseWindow->GetMainWidget();
     }
   }
-  if (!parentWidget && mView && mView->GetParent()) {
-    parentWidget = mView->GetParent()->GetNearestWidget(nullptr);
+  if (!parentWidget) {
+    parentWidget = GetParent()->GetNearestWidget();
   }
   return parentWidget.forget();
 }
 
-nsresult nsMenuPopupFrame::CreateWidgetForView(nsView* aView) {
+void nsMenuPopupFrame::CreateWidget() {
   // Create a widget for ourselves.
   widget::InitData widgetData;
   widgetData.mWindowType = widget::WindowType::Popup;
@@ -326,21 +321,39 @@ nsresult nsMenuPopupFrame::CreateWidgetForView(nsView* aView) {
 
   nsCOMPtr<nsIWidget> parentWidget = ComputeParentWidget();
   if (NS_WARN_IF(!parentWidget)) {
-    return NS_ERROR_FAILURE;
+    return;
   }
 
-  nsresult rv = aView->CreateWidgetForPopup(&widgetData, parentWidget);
-  if (NS_FAILED(rv)) {
-    return rv;
+  mWidget = parentWidget->CreateChild(CalcWidgetBounds(), widgetData);
+  if (NS_WARN_IF(!mWidget)) {
+    return;
   }
-
-  nsIWidget* widget = aView->GetWidget();
-  MOZ_ASSERT(widget->GetParent() == parentWidget);
-
-  widget->SetTransparencyMode(mode);
+  mWidget->SetWidgetListener(this);
+  // TODO(emilio): Make all widgets look at widgetData.mTransparencyMode
+  // (maybe in BaseCreate?) then remove this call.
+  mWidget->SetTransparencyMode(mode);
   PropagateStyleToWidget();
+}
 
-  return NS_OK;
+LayoutDeviceIntRect nsMenuPopupFrame::CalcWidgetBounds() const {
+  return nsView::CalcWidgetBounds(
+      GetRect(), PresContext()->AppUnitsPerDevPixel(),
+      PresShell()->GetViewManager()->GetRootView(), nullptr,
+      widget::WindowType::Popup,
+      nsLayoutUtils::GetFrameTransparency(this, this));
+}
+
+void nsMenuPopupFrame::DestroyWidget() {
+  RefPtr widget = mWidget.forget();
+  if (!widget) {
+    return;
+  }
+  // Widget's WebRender resources needs to be cleared before creating new
+  // widget.
+  widget->ClearCachedWebrenderResources();
+  widget->SetWidgetListener(nullptr);
+  NS_DispatchToMainThread(
+      NewRunnableMethod("DestroyWidget", widget, &nsIWidget::Destroy));
 }
 
 void nsMenuPopupFrame::PropagateStyleToWidget(WidgetStyleFlags aFlags) const {
@@ -500,6 +513,9 @@ void nsMenuPopupFrame::DidSetComputedStyle(ComputedStyle* aOldStyle) {
 
 nscoord nsMenuPopupFrame::IntrinsicISize(const IntrinsicSizeInput& aInput,
                                          IntrinsicISizeType aType) {
+  if (CanSkipLayout()) {
+    return 0;
+  }
   nscoord iSize = nsBlockFrame::IntrinsicISize(aInput, aType);
   if (!ShouldExpandToInflowParentOrAnchor()) {
     return iSize;
@@ -566,6 +582,18 @@ void nsMenuPopupFrame::EnsureActiveMenuListItemIsVisible() {
       ScrollFlags::ScrollOverflowHidden | ScrollFlags::ScrollFirstAncestorOnly);
 }
 
+bool nsMenuPopupFrame::CanSkipLayout() const {
+  // If the popup is not open, only do layout while showing or if we're a
+  // menulist.
+  //
+  // The later is needed because the SelectParent code wants to limit the height
+  // of the popup before opening it.
+  //
+  // TODO(emilio): We should consider adding a way to do that more reliably
+  // instead, but this preserves existing behavior.
+  return !IsVisibleOrShowing() && !IsMenuList();
+}
+
 void nsMenuPopupFrame::LayoutPopup(nsPresContext* aPresContext,
                                    ReflowOutput& aDesiredSize,
                                    const ReflowInput& aReflowInput,
@@ -577,21 +605,9 @@ void nsMenuPopupFrame::LayoutPopup(nsPresContext* aPresContext,
   SchedulePaint();
 
   const bool isOpen = IsOpen();
-  if (!isOpen) {
-    // If the popup is not open, only do layout while showing or if we're a
-    // menulist.
-    //
-    // This is needed because the SelectParent code wants to limit the height of
-    // the popup before opening it.
-    //
-    // TODO(emilio): We should consider adding a way to do that more reliably
-    // instead, but this preserves existing behavior.
-    const bool needsLayout = mPopupState == ePopupShowing ||
-                             mPopupState == ePopupPositioning || IsMenuList();
-    if (!needsLayout) {
-      RemoveStateBits(NS_FRAME_FIRST_REFLOW);
-      return;
-    }
+  if (CanSkipLayout()) {
+    RemoveStateBits(NS_FRAME_FIRST_REFLOW);
+    return;
   }
 
   // Do a first reflow, with all our content, in order to find our preferred
@@ -644,20 +660,17 @@ void nsMenuPopupFrame::LayoutPopup(nsPresContext* aPresContext,
                          aStatus);
   }
 
-  // Set our size, since nsAbsoluteContainingBlock won't.
+  if (mIsOpenChanged || !mRect.IsEqualEdges(constraints.mUsedRect)) {
+    SchedulePendingWidgetMoveResize();
+  }
+
+  // Set our size, since AbsoluteContainingBlock won't.
   SetRect(constraints.mUsedRect);
 
-  nsView* view = GetView();
   if (isOpen) {
-    nsViewManager* viewManager = view->GetViewManager();
-    viewManager->ResizeView(view,
-                            nsRect(nsPoint(), constraints.mUsedRect.Size()));
     if (mPopupState == ePopupOpening) {
       mPopupState = ePopupVisible;
     }
-
-    viewManager->SetViewVisibility(view, ViewVisibility::Show);
-    SyncFrameViewProperties(view);
   }
 
   // Perform our move now. That will position the view and so on.
@@ -962,6 +975,14 @@ void nsMenuPopupFrame::InitializePopupAsNativeContextMenu(
   mAnchorType = MenuPopupAnchorType::Point;
   mPositionedOffset = 0;
   mPositionedByMoveToRect = false;
+  // Native context menus don't call PrepareWidget(), so if we have a widget
+  // already (which generally should only be possible on tests, since
+  // otherwise we shouldn't ever mix native / non-native for the same popup) we
+  // should destroy it now.
+  if (mExpirationState.IsTracked()) {
+    PopupExpirationTracker::Get()->RemoveObject(this);
+  }
+  DestroyWidget();
 }
 
 void nsMenuPopupFrame::InitializePopupAtRect(nsIContent* aTriggerContent,
@@ -977,6 +998,7 @@ void nsMenuPopupFrame::ShowPopup(bool aIsContextMenu) {
   mIsContextMenu = aIsContextMenu;
 
   InvalidateFrameSubtree();
+  SchedulePendingWidgetMoveResize();
 
   if (mPopupState == ePopupShowing || mPopupState == ePopupPositioning) {
     mPopupState = ePopupOpening;
@@ -1066,17 +1088,15 @@ void nsMenuPopupFrame::HidePopup(bool aDeselectMenu, nsPopupState aNewState,
   mHFlip = mVFlip = false;
   mConstrainedByLayout = false;
 
-  if (auto* widget = GetWidget()) {
+  RefPtr widget = GetWidget();
+  if (widget) {
     widget->ClearCachedWebrenderResources();
     if (!aFromFrameDestruction && !ShouldHaveWidgetWhenHidden()) {
       PopupExpirationTracker::GetOrCreate().AddObject(this);
     }
   }
 
-  nsView* view = GetView();
-  nsViewManager* viewManager = view->GetViewManager();
-  viewManager->SetViewVisibility(view, ViewVisibility::Hide);
-
+  ClearPendingWidgetMoveResize();
   RefPtr popup = &PopupElement();
   // XXX, bug 137033, In Windows, if mouse is outside the window when the
   // menupopup closes, no mouse_enter/mouse_exit event will be fired to clear
@@ -1088,6 +1108,24 @@ void nsMenuPopupFrame::HidePopup(bool aDeselectMenu, nsPopupState aNewState,
     esm->SetContentState(nullptr, dom::ElementState::HOVER);
   }
   popup->PopupClosed(aDeselectMenu);
+
+  if (widget) {
+    nsContentUtils::AddScriptRunner(
+        NS_NewRunnableFunction("HideWidget", [widget = std::move(widget)] {
+          auto* frame = widget->GetPopupFrame();
+          if (!frame || !frame->IsVisibleOrShowing()) {
+            widget->Show(false);
+          }
+        }));
+  }
+}
+
+void nsMenuPopupFrame::SchedulePendingWidgetMoveResize() {
+  if (mPendingWidgetMoveResize) {
+    return;
+  }
+  mPendingWidgetMoveResize = true;
+  SchedulePaint();
 }
 
 nsPoint nsMenuPopupFrame::AdjustPositionForAnchorAlign(
@@ -1453,9 +1491,6 @@ auto nsMenuPopupFrame::GetRects(const nsSize& aPrefSize) const -> Rects {
 
   nsPresContext* pc = PresContext();
   nsIFrame* rootFrame = pc->PresShell()->GetRootFrame();
-  NS_ASSERTION(rootFrame->GetView() && GetView() &&
-                   rootFrame->GetView() == GetView()->GetParent(),
-               "rootFrame's view is not our view's parent???");
 
   // Indicators of whether the popup should be flipped or resized.
   FlipStyle hFlip = FlipStyle::None;
@@ -1555,10 +1590,7 @@ auto nsMenuPopupFrame::GetRects(const nsSize& aPrefSize) const -> Rects {
 
   const int32_t a2d = pc->AppUnitsPerDevPixel();
 
-  nsView* view = GetView();
-  NS_ASSERTION(view, "popup with no view");
-
-  nsIWidget* widget = view->GetWidget();
+  nsIWidget* widget = mWidget;
 
   // If a panel has flip="none", don't constrain or flip it.
   // Also, always do this for content shells, so that the popup doesn't extend
@@ -1719,13 +1751,14 @@ void nsMenuPopupFrame::SetPopupPosition(bool aIsMove) {
 void nsMenuPopupFrame::PerformMove(const Rects& aRects) {
   auto* ps = PresShell();
 
-  // We're just moving, sync frame position and offset as needed.
-  ps->GetViewManager()->MoveViewTo(GetView(), aRects.mViewPoint.x,
-                                   aRects.mViewPoint.y);
-
   // Now that we've positioned the view, sync up the frame's origin.
-  nsBlockFrame::SetPosition(aRects.mViewPoint -
-                            GetParent()->GetOffsetTo(ps->GetRootFrame()));
+  const nsPoint oldPos = mRect.TopLeft();
+  const nsPoint newPos =
+      aRects.mViewPoint - GetParent()->GetOffsetTo(ps->GetRootFrame());
+  nsBlockFrame::SetPosition(newPos);
+  if (oldPos != newPos) {
+    SchedulePendingWidgetMoveResize();
+  }
 
   // If the popup is in the positioned state or if it is shown and the position
   // or size changed, dispatch a popuppositioned event if the popup wants it.
@@ -2078,15 +2111,13 @@ XULButtonElement* nsMenuPopupFrame::FindMenuWithShortcut(
   return nullptr;
 }
 
-nsIWidget* nsMenuPopupFrame::GetWidget() const {
-  return mView ? mView->GetWidget() : nullptr;
-}
+nsIWidget* nsMenuPopupFrame::GetWidget() const { return mWidget.get(); }
 
 // helpers /////////////////////////////////////////////////////////////
 
 nsresult nsMenuPopupFrame::AttributeChanged(int32_t aNameSpaceID,
                                             nsAtom* aAttribute,
-                                            int32_t aModType)
+                                            AttrModType aModType)
 
 {
   nsresult rv =
@@ -2113,15 +2144,11 @@ nsresult nsMenuPopupFrame::AttributeChanged(int32_t aNameSpaceID,
 
   if (aAttribute == nsGkAtoms::label) {
     // set the label for the titlebar
-    nsView* view = GetView();
-    if (view) {
-      nsIWidget* widget = view->GetWidget();
-      if (widget) {
-        nsAutoString title;
-        mContent->AsElement()->GetAttr(nsGkAtoms::label, title);
-        if (!title.IsEmpty()) {
-          widget->SetTitle(title);
-        }
+    if (nsIWidget* widget = GetWidget()) {
+      nsAutoString title;
+      mContent->AsElement()->GetAttr(nsGkAtoms::label, title);
+      if (!title.IsEmpty()) {
+        widget->SetTitle(title);
       }
     }
   } else if (aAttribute == nsGkAtoms::ignorekeys) {
@@ -2159,7 +2186,6 @@ void nsMenuPopupFrame::Destroy(DestroyContext& aContext) {
   // but alas, also pre-existing.
   HidePopup(/* aDeselectMenu = */ false, ePopupClosed,
             /* aFromFrameDestruction = */ true);
-
   if (mExpirationState.IsTracked()) {
     PopupExpirationTracker::Get()->RemoveObject(this);
   }
@@ -2168,6 +2194,7 @@ void nsMenuPopupFrame::Destroy(DestroyContext& aContext) {
     pm->PopupDestroyed(this);
   }
 
+  DestroyWidget();
   nsBlockFrame::Destroy(aContext);
 }
 
@@ -2211,9 +2238,7 @@ void nsMenuPopupFrame::DestroyWidgetIfNeeded() {
     MOZ_ASSERT_UNREACHABLE("Shouldn't be tracked while visible");
     return;
   }
-  if (auto* view = GetView()) {
-    view->DestroyWidget();
-  }
+  DestroyWidget();
 }
 
 void nsMenuPopupFrame::MoveTo(const CSSPoint& aPos, bool aUpdateAttrs,
@@ -2341,37 +2366,6 @@ int8_t nsMenuPopupFrame::GetAlignmentPosition() const {
   return position;
 }
 
-/**
- * KEEP THIS IN SYNC WITH nsIFrame::CreateView
- * as much as possible. Until we get rid of views finally...
- */
-void nsMenuPopupFrame::CreatePopupView() {
-  if (HasView()) {
-    return;
-  }
-
-  nsViewManager* viewManager = PresContext()->GetPresShell()->GetViewManager();
-  NS_ASSERTION(nullptr != viewManager, "null view manager");
-
-  // Create a view
-  nsView* parentView = viewManager->GetRootView();
-  auto visibility = ViewVisibility::Hide;
-
-  NS_ASSERTION(parentView, "no parent view");
-
-  // Create a view
-  nsView* view = viewManager->CreateView(GetRect(), parentView, visibility);
-  // XXX put view last in document order until we can do better
-  viewManager->InsertChild(parentView, view, nullptr, true);
-
-  // Remember our view
-  SetView(view);
-
-  NS_FRAME_LOG(
-      NS_FRAME_TRACE_CALLS,
-      ("nsMenuPopupFrame::CreatePopupView: frame=%p view=%p", this, view));
-}
-
 bool nsMenuPopupFrame::ShouldFollowAnchor() const {
   if (mAnchorType != MenuPopupAnchorType::Node || !mAnchorContent) {
     return false;
@@ -2458,8 +2452,7 @@ void nsMenuPopupFrame::CheckForAnchorChange(nsRect& aRect) {
   }
 
   if (shouldHide) {
-    nsXULPopupManager* pm = nsXULPopupManager::GetInstance();
-    if (pm) {
+    if (nsXULPopupManager* pm = nsXULPopupManager::GetInstance()) {
       // As the caller will be iterating over the open popups, hide
       // asyncronously.
       pm->HidePopup(mContent->AsElement(),
@@ -2477,3 +2470,115 @@ void nsMenuPopupFrame::CheckForAnchorChange(nsRect& aRect) {
     SetPopupPosition(true);
   }
 }
+
+bool nsMenuPopupFrame::WindowMoved(nsIWidget* aWidget, int32_t aX, int32_t aY,
+                                   ByMoveToRect aByMoveToRect) {
+  MOZ_ASSERT(aWidget == mWidget);
+
+  if (!IsVisibleOrShowing()) {
+    return true;
+  }
+
+  LayoutDeviceIntPoint point(aX, aY);
+
+  // Don't do anything if the popup is already at the specified location. This
+  // prevents recursive calls when a popup is positioned.
+  LayoutDeviceIntRect curDevBounds = CalcWidgetBounds();
+  if (curDevBounds.TopLeft() == point &&
+      aWidget->GetClientOffset() == GetLastClientOffset()) {
+    return true;
+  }
+
+  // Update the popup's position using SetPopupPosition if the popup is
+  // anchored and at the parent level as these maintain their position
+  // relative to the parent window (except if positioned by move to rect, in
+  // which case we better make sure that layout matches that). Otherwise, just
+  // update the popup to the specified screen coordinates.
+  if (IsAnchored() && GetPopupLevel() == widget::PopupLevel::Parent &&
+      aByMoveToRect == ByMoveToRect::No) {
+    SetPopupPosition(true);
+  } else {
+    CSSPoint cssPos = point / PresContext()->CSSToDevPixelScale();
+    MoveTo(cssPos, false, aByMoveToRect == ByMoveToRect::Yes);
+  }
+  return true;
+}
+
+bool nsMenuPopupFrame::WindowResized(nsIWidget* aWidget, int32_t aWidth,
+                                     int32_t aHeight) {
+  MOZ_ASSERT(aWidget == mWidget);
+  if (!IsVisibleOrShowing()) {
+    return true;
+  }
+
+  LayoutDeviceIntSize size(aWidth, aHeight);
+  const LayoutDeviceIntRect curDevBounds = CalcWidgetBounds();
+  // If the size is what we think it is, we have nothing to do.
+  if (curDevBounds.Size() == size) {
+    return true;
+  }
+
+  RefPtr<Element> popup = &PopupElement();
+
+  // Only set the width and height if the popup already has these attributes.
+  if (!popup->HasAttr(nsGkAtoms::width) || !popup->HasAttr(nsGkAtoms::height)) {
+    return true;
+  }
+
+  // The size is different. Convert the actual size to css pixels and store it
+  // as 'width' and 'height' attributes on the popup.
+  nsPresContext* presContext = PresContext();
+
+  CSSIntSize newCSS(presContext->DevPixelsToIntCSSPixels(size.width),
+                    presContext->DevPixelsToIntCSSPixels(size.height));
+
+  nsAutoString width, height;
+  width.AppendInt(newCSS.width);
+  height.AppendInt(newCSS.height);
+  popup->SetAttr(kNameSpaceID_None, nsGkAtoms::width, width, true);
+  popup->SetAttr(kNameSpaceID_None, nsGkAtoms::height, height, true);
+  return true;
+}
+
+bool nsMenuPopupFrame::RequestWindowClose(nsIWidget* aWidget) {
+  MOZ_ASSERT(aWidget == mWidget);
+  if (nsXULPopupManager* pm = nsXULPopupManager::GetInstance()) {
+    pm->HidePopup(&PopupElement(), {HidePopupOption::DeselectMenu});
+    return true;
+  }
+  return false;
+}
+
+nsEventStatus nsMenuPopupFrame::HandleEvent(mozilla::WidgetGUIEvent* aEvent,
+                                            bool aUseAttachedEvents) {
+  MOZ_ASSERT(aEvent->mWidget);
+  MOZ_ASSERT(aEvent->mWidget == mWidget);
+  nsEventStatus status = nsEventStatus_eIgnore;
+  RefPtr ps = PresShell();
+  ps->HandleEvent(this, aEvent, false, &status);
+  return status;
+}
+
+bool nsMenuPopupFrame::PaintWindow(nsIWidget* aWidget, LayoutDeviceIntRegion) {
+  MOZ_ASSERT(aWidget == mWidget);
+  nsAutoScriptBlocker scriptBlocker;
+  RefPtr ps = PresShell();
+  RefPtr<WindowRenderer> renderer = aWidget->GetWindowRenderer();
+  if (!renderer->NeedsWidgetInvalidation()) {
+    renderer->FlushRendering(wr::RenderReasons::WIDGET);
+  } else {
+    ps->SyncPaintFallback(this, renderer);
+  }
+  return true;
+}
+
+void nsMenuPopupFrame::DidCompositeWindow(
+    mozilla::layers::TransactionId aTransactionId,
+    const TimeStamp& aCompositeStart, const TimeStamp& aCompositeEnd) {
+  RefPtr rootPc = PresContext()->GetRootPresContext();
+  if (!rootPc) {
+    return;
+  }
+  nsAutoScriptBlocker scriptBlocker;
+  rootPc->NotifyDidPaintForSubtree(aTransactionId, aCompositeEnd);
+}
