@@ -10,19 +10,25 @@ use crate::data::ElementData;
 use crate::dom::{TElement, TNode};
 use crate::invalidation::element::element_wrapper::{ElementSnapshot, ElementWrapper};
 use crate::invalidation::element::invalidation_map::*;
-use crate::invalidation::element::invalidator::{DescendantInvalidationLists, InvalidationVector};
+use crate::invalidation::element::invalidator::{
+    any_next_has_scope_in_negation, note_scope_dependency_force_at_subject,
+    DescendantInvalidationLists, InvalidationVector, SiblingTraversalMap,
+};
 use crate::invalidation::element::invalidator::{Invalidation, InvalidationProcessor};
 use crate::invalidation::element::restyle_hints::RestyleHint;
 use crate::selector_map::SelectorMap;
 use crate::selector_parser::Snapshot;
 use crate::stylesheets::origin::OriginSet;
+use crate::values::AtomIdent;
 use crate::{Atom, WeakAtom};
 use dom::ElementState;
 use selectors::attr::CaseSensitivity;
+use selectors::kleene_value::KleeneValue;
 use selectors::matching::{
-    matches_selector, MatchingForInvalidation, MatchingContext, MatchingMode,
-    NeedsSelectorFlags, SelectorCaches, VisitedHandlingMode,
+    matches_selector_kleene, IncludeStartingStyle, MatchingContext, MatchingForInvalidation,
+    MatchingMode, NeedsSelectorFlags, SelectorCaches, VisitedHandlingMode,
 };
+use selectors::OpaqueElement;
 use smallvec::SmallVec;
 
 /// The collector implementation.
@@ -39,6 +45,8 @@ where
     added_id: Option<&'a WeakAtom>,
     classes_removed: &'a SmallVec<[Atom; 8]>,
     classes_added: &'a SmallVec<[Atom; 8]>,
+    custom_states_removed: &'a SmallVec<[AtomIdent; 8]>,
+    custom_states_added: &'a SmallVec<[AtomIdent; 8]>,
     state_changes: ElementState,
     descendant_invalidations: &'a mut DescendantInvalidationLists<'selectors>,
     sibling_invalidations: &'a mut InvalidationVector<'selectors>,
@@ -52,6 +60,7 @@ pub struct StateAndAttrInvalidationProcessor<'a, 'b: 'a, E: TElement> {
     element: E,
     data: &'a mut ElementData,
     matching_context: MatchingContext<'a, E::Impl>,
+    traversal_map: SiblingTraversalMap<E>,
 }
 
 impl<'a, 'b: 'a, E: TElement + 'b> StateAndAttrInvalidationProcessor<'a, 'b, E> {
@@ -67,6 +76,7 @@ impl<'a, 'b: 'a, E: TElement + 'b> StateAndAttrInvalidationProcessor<'a, 'b, E>
             None,
             selector_caches,
             VisitedHandlingMode::AllLinksVisitedAndUnvisited,
+            IncludeStartingStyle::No,
             shared_context.quirks_mode(),
             NeedsSelectorFlags::No,
             MatchingForInvalidation::Yes,
@@ -77,6 +87,7 @@ impl<'a, 'b: 'a, E: TElement + 'b> StateAndAttrInvalidationProcessor<'a, 'b, E>
             element,
             data,
             matching_context,
+            traversal_map: SiblingTraversalMap::default(),
         }
     }
 }
@@ -88,28 +99,41 @@ pub fn check_dependency<E, W>(
     element: &E,
     wrapper: &W,
     context: &mut MatchingContext<'_, E::Impl>,
+    scope: Option<OpaqueElement>,
 ) -> bool
 where
     E: TElement,
     W: selectors::Element<Impl = E::Impl>,
 {
-    let matches_now = matches_selector(
-        &dependency.selector,
-        dependency.selector_offset,
-        None,
-        element,
-        context,
-    );
+    context.for_invalidation_comparison(|context| {
+        context.nest_for_scope_condition(scope, |context| {
+            let matches_now = matches_selector_kleene(
+                &dependency.selector,
+                dependency.selector_offset,
+                None,
+                element,
+                context,
+            );
 
-    let matched_then = matches_selector(
-        &dependency.selector,
-        dependency.selector_offset,
-        None,
-        wrapper,
-        context,
-    );
+            // If the previous dependency was a scope dependency (i.e. by `scope` is set),
+            // possible change in scope element is encapsulated in `:scope`, whose
+            // matching value will not change. We instead check that the change in scope
+            // element can propagate (i.e. This selector matches).
+            if scope.is_some() && matches_now != KleeneValue::False {
+                return true;
+            }
 
-    matched_then != matches_now
+            let matched_then = matches_selector_kleene(
+                &dependency.selector,
+                dependency.selector_offset,
+                None,
+                wrapper,
+                context,
+            );
+
+            matched_then != matches_now || matches_now == KleeneValue::Unknown
+        })
+    })
 }
 
 /// Whether we should process the descendants of a given element for style
@@ -194,7 +218,7 @@ where
     }
 }
 
-impl<'a, 'b: 'a, E: 'a> InvalidationProcessor<'a, E>
+impl<'a, 'b: 'a, E: 'a> InvalidationProcessor<'a, 'a, E>
     for StateAndAttrInvalidationProcessor<'a, 'b, E>
 where
     E: TElement,
@@ -206,18 +230,33 @@ where
         true
     }
 
-    fn check_outer_dependency(&mut self, dependency: &Dependency, element: E) -> bool {
+    fn check_outer_dependency(
+        &mut self,
+        dependency: &Dependency,
+        element: E,
+        scope: Option<OpaqueElement>,
+    ) -> bool {
         // We cannot assert about `element` having a snapshot here (in fact it
         // most likely won't), because it may be an arbitrary descendant or
         // later-sibling of the element we started invalidating with.
         let wrapper = ElementWrapper::new(element, &*self.shared_context.snapshot_map);
-        check_dependency(dependency, &element, &wrapper, &mut self.matching_context)
+        check_dependency(
+            dependency,
+            &element,
+            &wrapper,
+            &mut self.matching_context,
+            scope,
+        )
     }
 
     fn matching_context(&mut self) -> &mut MatchingContext<'a, E::Impl> {
         &mut self.matching_context
     }
 
+    fn sibling_traversal_map(&self) -> &SiblingTraversalMap<E> {
+        &self.traversal_map
+    }
+
     fn collect_invalidations(
         &mut self,
         element: E,
@@ -231,9 +270,11 @@ where
         let wrapper = ElementWrapper::new(element, &*self.shared_context.snapshot_map);
 
         let state_changes = wrapper.state_changes();
-        let snapshot = wrapper.snapshot().expect("has_snapshot lied");
+        let Some(snapshot) = wrapper.snapshot() else {
+            return false;
+        };
 
-        if !snapshot.has_attrs() && state_changes.is_empty() {
+        if !snapshot.has_attrs() && !snapshot.has_custom_states() && state_changes.is_empty() {
             return false;
         }
 
@@ -254,6 +295,21 @@ where
             })
         }
 
+        let mut custom_states_removed = SmallVec::<[AtomIdent; 8]>::new();
+        let mut custom_states_added = SmallVec::<[AtomIdent; 8]>::new();
+        if snapshot.has_custom_states() {
+            snapshot.each_custom_state(|s| {
+                if !element.has_custom_state(s) {
+                    custom_states_removed.push(s.clone())
+                }
+            });
+            element.each_custom_state(|s| {
+                if !snapshot.has_custom_state(s) {
+                    custom_states_added.push(s.clone())
+                }
+            })
+        }
+
         let mut id_removed = None;
         let mut id_added = None;
         if snapshot.id_changed() {
@@ -316,6 +372,8 @@ where
                 added_id: id_added,
                 classes_removed: &classes_removed,
                 classes_added: &classes_added,
+                custom_states_removed: &custom_states_removed,
+                custom_states_added: &custom_states_added,
                 descendant_invalidations,
                 sibling_invalidations,
                 invalidates_self: false,
@@ -411,7 +469,7 @@ where
         if let Some(ref id) = removed_id {
             if let Some(deps) = map.id_to_selector.get(id, quirks_mode) {
                 for dep in deps {
-                    self.scan_dependency(dep);
+                    self.scan_dependency(dep, false);
                 }
             }
         }
@@ -420,7 +478,7 @@ where
         if let Some(ref id) = added_id {
             if let Some(deps) = map.id_to_selector.get(id, quirks_mode) {
                 for dep in deps {
-                    self.scan_dependency(dep);
+                    self.scan_dependency(dep, false);
                 }
             }
         }
@@ -428,7 +486,19 @@ where
         for class in self.classes_added.iter().chain(self.classes_removed.iter()) {
             if let Some(deps) = map.class_to_selector.get(class, quirks_mode) {
                 for dep in deps {
-                    self.scan_dependency(dep);
+                    self.scan_dependency(dep, false);
+                }
+            }
+        }
+
+        for state in self
+            .custom_states_added
+            .iter()
+            .chain(self.custom_states_removed.iter())
+        {
+            if let Some(deps) = map.custom_state_affecting_selectors.get(state) {
+                for dep in deps {
+                    self.scan_dependency(dep, false);
                 }
             }
         }
@@ -436,7 +506,7 @@ where
         self.snapshot.each_attr_changed(|attribute| {
             if let Some(deps) = map.other_attribute_affecting_selectors.get(attribute) {
                 for dep in deps {
-                    self.scan_dependency(dep);
+                    self.scan_dependency(dep, false);
                 }
             }
         });
@@ -458,7 +528,7 @@ where
                 if !dependency.state.intersects(self.state_changes) {
                     return true;
                 }
-                self.scan_dependency(&dependency.dep);
+                self.scan_dependency(&dependency.dep, false);
                 true
             },
         );
@@ -466,16 +536,24 @@ where
 
     /// Check whether a dependency should be taken into account.
     #[inline]
-    fn check_dependency(&mut self, dependency: &Dependency) -> bool {
+    fn check_dependency(&mut self, dependency: &Dependency, set_scope: bool) -> bool {
         check_dependency(
             dependency,
             &self.element,
             &self.wrapper,
             &mut self.matching_context,
+            set_scope.then(|| self.element.opaque()),
         )
     }
 
-    fn scan_dependency(&mut self, dependency: &'selectors Dependency) {
+    fn scan_dependency(&mut self, dependency: &'selectors Dependency, set_scope: bool) {
+        debug_assert!(
+            matches!(
+                dependency.invalidation_kind(),
+                DependencyInvalidationKind::Normal(_) | DependencyInvalidationKind::Scope(_)
+            ),
+            "Found unexpected dependency invalidation kind"
+        );
         debug!(
             "TreeStyleInvalidator::scan_dependency({:?}, {:?})",
             self.element, dependency
@@ -485,69 +563,177 @@ where
             return;
         }
 
-        if self.check_dependency(dependency) {
-            return self.note_dependency(dependency);
+        if self.check_dependency(dependency, set_scope) {
+            return self.note_dependency(dependency, set_scope);
         }
     }
 
-    fn note_dependency(&mut self, dependency: &'selectors Dependency) {
+    fn note_dependency(&mut self, dependency: &'selectors Dependency, set_scope: bool) {
         debug_assert!(self.dependency_may_be_relevant(dependency));
-
         let invalidation_kind = dependency.invalidation_kind();
-        if matches!(invalidation_kind, DependencyInvalidationKind::Element) {
-            if let Some(ref parent) = dependency.parent {
+        if matches!(
+            invalidation_kind,
+            DependencyInvalidationKind::Normal(NormalDependencyInvalidationKind::Element)
+        ) {
+            if let Some(ref next) = dependency.next {
                 // We know something changed in the inner selector, go outwards
                 // now.
-                self.scan_dependency(parent);
+                self.scan_dependency(&next.as_ref().slice()[0], set_scope);
             } else {
                 self.invalidates_self = true;
             }
             return;
         }
 
+        if let DependencyInvalidationKind::Scope(scope_kind) = invalidation_kind {
+            if scope_kind == ScopeDependencyInvalidationKind::ImplicitScope {
+                if let Some(ref next) = dependency.next {
+                    // When we reach an implicit scope dependency, we know there's an
+                    // element matching that implicit scope somewhere in the descendant.
+                    // We need to go find it so that we can continue the invalidation from
+                    // its next dependencies.
+                    for dep in next.as_ref().slice() {
+                        let invalidation = Invalidation::new_always_effective_for_next_descendant(
+                            dep,
+                            self.matching_context.current_host.clone(),
+                            self.matching_context.scope_element,
+                        );
+
+                        self.descendant_invalidations
+                            .dom_descendants
+                            .push(invalidation);
+                    }
+                    return;
+                }
+            }
+
+            if dependency.selector.is_rightmost(dependency.selector_offset) {
+                let force_add = any_next_has_scope_in_negation(dependency);
+                if scope_kind == ScopeDependencyInvalidationKind::ScopeEnd || force_add {
+                    let invalidations = note_scope_dependency_force_at_subject(
+                        dependency,
+                        self.matching_context.current_host.clone(),
+                        self.matching_context.scope_element,
+                        force_add,
+                    );
+                    for invalidation in invalidations {
+                        self.descendant_invalidations
+                            .dom_descendants
+                            .push(invalidation);
+                    }
+                    self.invalidates_self = true;
+                } else if let Some(ref next) = dependency.next {
+                    for dep in next.as_ref().slice() {
+                        self.scan_dependency(dep, true);
+                    }
+                }
+                return;
+            }
+        }
+
         debug_assert_ne!(dependency.selector_offset, 0);
         debug_assert_ne!(dependency.selector_offset, dependency.selector.len());
 
-        let invalidation =
-            Invalidation::new(&dependency, self.matching_context.current_host.clone());
+        let invalidation = Invalidation::new(
+            &dependency,
+            self.matching_context.current_host.clone(),
+            self.matching_context.scope_element.clone(),
+        );
 
-        match invalidation_kind {
-            DependencyInvalidationKind::Element => unreachable!(),
-            DependencyInvalidationKind::ElementAndDescendants => {
-                self.invalidates_self = true;
-                self.descendant_invalidations
-                    .dom_descendants
-                    .push(invalidation);
-            },
-            DependencyInvalidationKind::Descendants => {
-                self.descendant_invalidations
-                    .dom_descendants
-                    .push(invalidation);
-            },
-            DependencyInvalidationKind::Siblings => {
-                self.sibling_invalidations.push(invalidation);
-            },
-            DependencyInvalidationKind::Parts => {
-                self.descendant_invalidations.parts.push(invalidation);
-            },
-            DependencyInvalidationKind::SlottedElements => {
-                self.descendant_invalidations
-                    .slotted_descendants
-                    .push(invalidation);
-            },
-        }
+        self.invalidates_self |= push_invalidation(
+            invalidation,
+            invalidation_kind,
+            self.descendant_invalidations,
+            self.sibling_invalidations,
+        );
     }
 
     /// Returns whether `dependency` may cause us to invalidate the style of
     /// more elements than what we've already invalidated.
     fn dependency_may_be_relevant(&self, dependency: &Dependency) -> bool {
         match dependency.invalidation_kind() {
-            DependencyInvalidationKind::Element => !self.invalidates_self,
-            DependencyInvalidationKind::SlottedElements => self.element.is_html_slot_element(),
-            DependencyInvalidationKind::Parts => self.element.shadow_root().is_some(),
-            DependencyInvalidationKind::ElementAndDescendants |
-            DependencyInvalidationKind::Siblings |
-            DependencyInvalidationKind::Descendants => true,
+            DependencyInvalidationKind::FullSelector | DependencyInvalidationKind::Relative(_) => {
+                unreachable!()
+            },
+            DependencyInvalidationKind::Scope(_) => true,
+            DependencyInvalidationKind::Normal(kind) => match kind {
+                NormalDependencyInvalidationKind::Element => !self.invalidates_self,
+                NormalDependencyInvalidationKind::SlottedElements => {
+                    self.element.is_html_slot_element()
+                },
+                NormalDependencyInvalidationKind::Parts => self.element.shadow_root().is_some(),
+                NormalDependencyInvalidationKind::ElementAndDescendants
+                | NormalDependencyInvalidationKind::Siblings
+                | NormalDependencyInvalidationKind::Descendants => true,
+            },
         }
     }
 }
+
+pub(crate) fn push_invalidation<'a>(
+    invalidation: Invalidation<'a>,
+    invalidation_kind: DependencyInvalidationKind,
+    descendant_invalidations: &mut DescendantInvalidationLists<'a>,
+    sibling_invalidations: &mut InvalidationVector<'a>,
+) -> bool {
+    match invalidation_kind {
+        DependencyInvalidationKind::FullSelector => unreachable!(),
+        DependencyInvalidationKind::Relative(_) => unreachable!(),
+        DependencyInvalidationKind::Scope(_) => {
+            // Scope invalidation kind matters only upon reaching the subject.
+            // Examine the combinator to the right of the compound.
+            let combinator = invalidation.combinator_to_right();
+            if combinator.is_sibling() {
+                sibling_invalidations.push(invalidation);
+            } else {
+                descendant_invalidations.dom_descendants.push(invalidation);
+            }
+            true
+        },
+        DependencyInvalidationKind::Normal(kind) => match kind {
+            NormalDependencyInvalidationKind::Element => unreachable!(),
+            NormalDependencyInvalidationKind::ElementAndDescendants => {
+                descendant_invalidations.dom_descendants.push(invalidation);
+                true
+            },
+            NormalDependencyInvalidationKind::Descendants => {
+                descendant_invalidations.dom_descendants.push(invalidation);
+                false
+            },
+            NormalDependencyInvalidationKind::Siblings => {
+                sibling_invalidations.push(invalidation);
+                false
+            },
+            NormalDependencyInvalidationKind::Parts => {
+                descendant_invalidations.parts.push(invalidation);
+                false
+            },
+            NormalDependencyInvalidationKind::SlottedElements => {
+                descendant_invalidations
+                    .slotted_descendants
+                    .push(invalidation);
+                false
+            },
+        },
+    }
+}
+
+pub(crate) fn dependency_may_be_relevant<E: TElement>(
+    dependency: &Dependency,
+    element: &E,
+    already_invalidated_self: bool,
+) -> bool {
+    match dependency.invalidation_kind() {
+        DependencyInvalidationKind::FullSelector => unreachable!(),
+        DependencyInvalidationKind::Relative(_) => unreachable!(),
+        DependencyInvalidationKind::Scope(_) => true,
+        DependencyInvalidationKind::Normal(kind) => match kind {
+            NormalDependencyInvalidationKind::Element => !already_invalidated_self,
+            NormalDependencyInvalidationKind::SlottedElements => element.is_html_slot_element(),
+            NormalDependencyInvalidationKind::Parts => element.shadow_root().is_some(),
+            NormalDependencyInvalidationKind::ElementAndDescendants
+            | NormalDependencyInvalidationKind::Siblings
+            | NormalDependencyInvalidationKind::Descendants => true,
+        },
+    }
+}
