@@ -6,38 +6,45 @@
 
 #![deny(missing_docs)]
 
-use super::generated::{
-    shorthands, AllShorthand, ComputedValues, LogicalGroupSet, LonghandIdSet,
-    NonCustomPropertyIdSet, PropertyDeclaration, PropertyDeclarationId, PropertyId, ShorthandId,
-    SourcePropertyDeclaration, SourcePropertyDeclarationDrain, SubpropertiesVec,
+use super::{
+    property_counts, AllShorthand, ComputedValues, LogicalGroupSet, LonghandIdSet,
+    LonghandIdSetIterator, NonCustomPropertyIdSet, PropertyDeclaration, PropertyDeclarationId,
+    PropertyId, ShorthandId, SourcePropertyDeclaration, SourcePropertyDeclarationDrain,
+    SubpropertiesVec,
 };
-use crate::applicable_declarations::CascadePriority;
 use crate::context::QuirksMode;
-use crate::custom_properties::{self, CustomPropertiesBuilder};
+use crate::custom_properties;
 use crate::error_reporting::{ContextualParseError, ParseErrorReporter};
 use crate::parser::ParserContext;
-use crate::properties::animated_properties::{AnimationValue, AnimationValueMap};
-use crate::stylist::Stylist;
-use crate::rule_tree::CascadeLevel;
+use crate::properties::{
+    animated_properties::{AnimationValue, AnimationValueMap},
+    StyleBuilder,
+};
+use crate::rule_cache::RuleCacheConditions;
 use crate::selector_map::PrecomputedHashSet;
 use crate::selector_parser::SelectorImpl;
 use crate::shared_lock::Locked;
-use crate::str::{CssString, CssStringWriter};
-use crate::stylesheets::{layer_rule::LayerOrder, CssRuleType, Origin, UrlExtraData};
+use crate::stylesheets::container_rule::ContainerSizeQuery;
+use crate::stylesheets::{CssRuleType, Origin, UrlExtraData};
+use crate::stylist::Stylist;
 use crate::values::computed::Context;
 use cssparser::{
     parse_important, AtRuleParser, CowRcStr, DeclarationParser, Delimiter, ParseErrorKind, Parser,
-    ParserInput, QualifiedRuleParser, RuleBodyItemParser, RuleBodyParser,
+    ParserInput, ParserState, QualifiedRuleParser, RuleBodyItemParser, RuleBodyParser,
+    SourceLocation,
 };
 use itertools::Itertools;
 use selectors::SelectorList;
 use servo_arc::Arc;
-use smallbitvec::{self, SmallBitVec};
+use smallbitvec::SmallBitVec;
 use smallvec::SmallVec;
 use std::fmt::{self, Write};
-use std::iter::{DoubleEndedIterator, Zip};
+use std::iter::Zip;
 use std::slice::Iter;
-use style_traits::{CssWriter, ParseError, ParsingMode, StyleParseErrorKind, ToCss};
+use style_traits::{
+    CssString, CssStringWriter, CssWriter, ParseError, ParsingMode, StyleParseErrorKind, ToCss,
+    TypedValue,
+};
 use thin_vec::ThinVec;
 
 /// A set of property declarations including animations and transitions.
@@ -108,14 +115,34 @@ impl Importance {
     }
 }
 
-#[derive(Clone, ToShmem, Default, MallocSizeOf)]
-struct PropertyDeclarationIdSet {
+/// A property-aware wrapper around reification results.
+///
+/// While `TypedValue` is property-agnostic, this enum represents the outcome
+/// of reifying a specific property inside a `PropertyDeclarationBlock`.
+#[derive(Clone, Debug)]
+pub enum PropertyTypedValue {
+    /// The property is not present in the declaration block.
+    None,
+
+    /// The property exists but cannot be expressed as a `TypedValue`.
+    /// Used for shorthands and other unrepresentable cases, which must be
+    /// exposed as `CSSUnsupportedValue` objects tied to the property.
+    Unsupported,
+
+    /// The property was successfully reified into a `TypedValue`.
+    Typed(TypedValue),
+}
+
+/// A set of properties.
+#[derive(Clone, Debug, ToShmem, Default, MallocSizeOf)]
+pub struct PropertyDeclarationIdSet {
     longhands: LonghandIdSet,
     custom: PrecomputedHashSet<custom_properties::Name>,
 }
 
 impl PropertyDeclarationIdSet {
-    fn insert(&mut self, id: PropertyDeclarationId) -> bool {
+    /// Add the given property to the set.
+    pub fn insert(&mut self, id: PropertyDeclarationId) -> bool {
         match id {
             PropertyDeclarationId::Longhand(id) => {
                 if self.longhands.contains(id) {
@@ -124,18 +151,20 @@ impl PropertyDeclarationIdSet {
                 self.longhands.insert(id);
                 return true;
             },
-            PropertyDeclarationId::Custom(name) => self.custom.insert(name.clone()),
+            PropertyDeclarationId::Custom(name) => self.custom.insert((*name).clone()),
         }
     }
 
-    fn contains(&self, id: PropertyDeclarationId) -> bool {
+    /// Return whether the given property is in the set.
+    pub fn contains(&self, id: PropertyDeclarationId) -> bool {
         match id {
             PropertyDeclarationId::Longhand(id) => self.longhands.contains(id),
             PropertyDeclarationId::Custom(name) => self.custom.contains(name),
         }
     }
 
-    fn remove(&mut self, id: PropertyDeclarationId) {
+    /// Remove the given property from the set.
+    pub fn remove(&mut self, id: PropertyDeclarationId) {
         match id {
             PropertyDeclarationId::Longhand(id) => self.longhands.remove(id),
             PropertyDeclarationId::Custom(name) => {
@@ -144,10 +173,74 @@ impl PropertyDeclarationIdSet {
         }
     }
 
-    fn clear(&mut self) {
+    /// Remove all properties from the set.
+    pub fn clear(&mut self) {
         self.longhands.clear();
         self.custom.clear();
     }
+
+    /// Returns whether the set is empty.
+    #[inline]
+    pub fn is_empty(&self) -> bool {
+        self.longhands.is_empty() && self.custom.is_empty()
+    }
+    /// Returns whether this set contains any reset longhand.
+    #[inline]
+    pub fn contains_any_reset(&self) -> bool {
+        self.longhands.contains_any_reset()
+    }
+
+    /// Returns whether this set contains all longhands in the specified set.
+    #[inline]
+    pub fn contains_all_longhands(&self, longhands: &LonghandIdSet) -> bool {
+        self.longhands.contains_all(longhands)
+    }
+
+    /// Returns whether this set contains all properties in the specified set.
+    #[inline]
+    pub fn contains_all(&self, properties: &PropertyDeclarationIdSet) -> bool {
+        if !self.longhands.contains_all(&properties.longhands) {
+            return false;
+        }
+        if properties.custom.len() > self.custom.len() {
+            return false;
+        }
+        properties
+            .custom
+            .iter()
+            .all(|item| self.custom.contains(item))
+    }
+
+    /// Iterate over the current property declaration id set.
+    pub fn iter(&self) -> PropertyDeclarationIdSetIterator<'_> {
+        PropertyDeclarationIdSetIterator {
+            longhands: self.longhands.iter(),
+            custom: self.custom.iter(),
+        }
+    }
+}
+
+/// An iterator over a set of longhand ids.
+pub struct PropertyDeclarationIdSetIterator<'a> {
+    longhands: LonghandIdSetIterator<'a>,
+    custom: std::collections::hash_set::Iter<'a, custom_properties::Name>,
+}
+
+impl<'a> Iterator for PropertyDeclarationIdSetIterator<'a> {
+    type Item = PropertyDeclarationId<'a>;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        // LonghandIdSetIterator's implementation always returns None
+        // after it did it once, so the code below will then continue
+        // to iterate over the custom properties.
+        match self.longhands.next() {
+            Some(id) => Some(PropertyDeclarationId::Longhand(id)),
+            None => match self.custom.next() {
+                Some(a) => Some(PropertyDeclarationId::Custom(a)),
+                None => None,
+            },
+        }
+    }
 }
 
 /// Overridden declarations are skipped.
@@ -231,23 +324,22 @@ impl<'a> DoubleEndedIterator for DeclarationImportanceIterator<'a> {
 pub struct AnimationValueIterator<'a, 'cx, 'cx_a: 'cx> {
     iter: DeclarationImportanceIterator<'a>,
     context: &'cx mut Context<'cx_a>,
+    style: &'a ComputedValues,
     default_values: &'a ComputedValues,
-    /// Custom properties in a keyframe if exists.
-    extra_custom_properties: Option<&'a Arc<crate::custom_properties::CustomPropertiesMap>>,
 }
 
 impl<'a, 'cx, 'cx_a: 'cx> AnimationValueIterator<'a, 'cx, 'cx_a> {
     fn new(
         declarations: &'a PropertyDeclarationBlock,
         context: &'cx mut Context<'cx_a>,
+        style: &'a ComputedValues,
         default_values: &'a ComputedValues,
-        extra_custom_properties: Option<&'a Arc<crate::custom_properties::CustomPropertiesMap>>,
     ) -> AnimationValueIterator<'a, 'cx, 'cx_a> {
         AnimationValueIterator {
             iter: declarations.declaration_importance_iter(),
             context,
+            style,
             default_values,
-            extra_custom_properties,
         }
     }
 }
@@ -266,7 +358,7 @@ impl<'a, 'cx, 'cx_a: 'cx> Iterator for AnimationValueIterator<'a, 'cx, 'cx_a> {
             let animation = AnimationValue::from_declaration(
                 decl,
                 &mut self.context,
-                self.extra_custom_properties,
+                self.style,
                 self.default_values,
             );
 
@@ -333,7 +425,7 @@ impl PropertyDeclarationBlock {
 
     /// Iterate over `(PropertyDeclaration, Importance)` pairs
     #[inline]
-    pub fn declaration_importance_iter(&self) -> DeclarationImportanceIterator {
+    pub fn declaration_importance_iter(&self) -> DeclarationImportanceIterator<'_> {
         DeclarationImportanceIterator::new(&self.declarations, &self.declarations_importance)
     }
 
@@ -352,10 +444,10 @@ impl PropertyDeclarationBlock {
     pub fn to_animation_value_iter<'a, 'cx, 'cx_a: 'cx>(
         &'a self,
         context: &'cx mut Context<'cx_a>,
+        style: &'a ComputedValues,
         default_values: &'a ComputedValues,
-        extra_custom_properties: Option<&'a Arc<crate::custom_properties::CustomPropertiesMap>>,
     ) -> AnimationValueIterator<'a, 'cx, 'cx_a> {
-        AnimationValueIterator::new(self, context, default_values, extra_custom_properties)
+        AnimationValueIterator::new(self, context, style, default_values)
     }
 
     /// Returns whether this block contains any declaration with `!important`.
@@ -376,11 +468,11 @@ impl PropertyDeclarationBlock {
         !self.declarations_importance.all_true()
     }
 
-    /// Returns a `LonghandIdSet` representing the properties that are changed in
+    /// Returns a `PropertyDeclarationIdSet` representing the properties that are changed in
     /// this block.
     #[inline]
-    pub fn longhands(&self) -> &LonghandIdSet {
-        &self.property_ids.longhands
+    pub fn property_ids(&self) -> &PropertyDeclarationIdSet {
+        &self.property_ids
     }
 
     /// Returns whether this block contains a declaration of a given property id.
@@ -392,7 +484,7 @@ impl PropertyDeclarationBlock {
     /// Returns whether this block contains any reset longhand.
     #[inline]
     pub fn contains_any_reset(&self) -> bool {
-        self.property_ids.longhands.contains_any_reset()
+        self.property_ids.contains_any_reset()
     }
 
     /// Get a declaration for a given property.
@@ -505,6 +597,32 @@ impl PropertyDeclarationBlock {
         }
     }
 
+    /// Find the value of the given property in this block and reify it
+    pub fn property_value_to_typed(&self, property: &PropertyId) -> PropertyTypedValue {
+        match property.as_shorthand() {
+            Ok(shorthand) => {
+                if shorthand
+                    .longhands()
+                    .all(|longhand| self.contains(PropertyDeclarationId::Longhand(longhand)))
+                {
+                    PropertyTypedValue::Unsupported
+                } else {
+                    PropertyTypedValue::None
+                }
+            },
+            Err(longhand_or_custom) => match self.get(longhand_or_custom) {
+                Some((value, _importance)) => {
+                    if let Some(typed_value) = value.to_typed() {
+                        PropertyTypedValue::Typed(typed_value)
+                    } else {
+                        PropertyTypedValue::Unsupported
+                    }
+                },
+                None => PropertyTypedValue::None,
+            },
+        }
+    }
+
     /// Adds or overrides the declaration for a given property in this block.
     ///
     /// See the documentation of `push` to see what impact `source` has when the
@@ -517,7 +635,7 @@ impl PropertyDeclarationBlock {
         let all_shorthand_len = match drain.all_shorthand {
             AllShorthand::NotSet => 0,
             AllShorthand::CSSWideKeyword(_) | AllShorthand::WithVariables(_) => {
-                shorthands::ALL_SHORTHAND_MAX_LEN
+                property_counts::ALL_SHORTHAND_EXPANDED
             },
         };
         let push_calls_count = drain.declarations.len() + all_shorthand_len;
@@ -594,8 +712,9 @@ impl PropertyDeclarationBlock {
                 .all_shorthand
                 .declarations()
                 .any(|decl| {
-                    !self.contains(decl.id()) ||
-                        self.declarations
+                    !self.contains(decl.id())
+                        || self
+                            .declarations
                             .iter()
                             .enumerate()
                             .find(|&(_, ref d)| d.id() == decl.id())
@@ -637,9 +756,9 @@ impl PropertyDeclarationBlock {
                                     }
                                     return DeclarationUpdate::UpdateInPlace { pos };
                                 }
-                                if !needs_append &&
-                                    id.logical_group() == Some(logical_group) &&
-                                    id.is_logical() != longhand_id.is_logical()
+                                if !needs_append
+                                    && id.logical_group() == Some(logical_group)
+                                    && id.is_logical() != longhand_id.is_logical()
                                 {
                                     needs_append = true;
                                 }
@@ -844,7 +963,6 @@ impl PropertyDeclarationBlock {
         property: &PropertyId,
         dest: &mut CssStringWriter,
         computed_values: Option<&ComputedValues>,
-        custom_properties_block: Option<&PropertyDeclarationBlock>,
         stylist: &Stylist,
     ) -> fmt::Result {
         if let Ok(shorthand) = property.as_shorthand() {
@@ -858,18 +976,23 @@ impl PropertyDeclarationBlock {
             None => return Err(fmt::Error),
         };
 
-        let custom_properties = if let Some(cv) = computed_values {
-            // If there are extra custom properties for this declaration block,
-            // factor them in too.
-            if let Some(block) = custom_properties_block {
-                // FIXME(emilio): This is not super-efficient here, and all this
-                // feels like a hack anyway...
-                block.cascade_custom_properties(cv.custom_properties(), stylist)
-            } else {
-                cv.custom_properties().cloned()
-            }
-        } else {
-            None
+        let mut rule_cache_conditions = RuleCacheConditions::default();
+        let mut context = Context::new(
+            StyleBuilder::new(
+                stylist.device(),
+                Some(stylist),
+                computed_values,
+                None,
+                None,
+                false,
+            ),
+            stylist.quirks_mode(),
+            &mut rule_cache_conditions,
+            ContainerSizeQuery::none(),
+        );
+
+        if let Some(cv) = computed_values {
+            context.builder.custom_properties = cv.custom_properties().clone();
         };
 
         match (declaration, computed_values) {
@@ -880,19 +1003,16 @@ impl PropertyDeclarationBlock {
             // getKeyframes() implementation for CSS animations, if
             // |computed_values| is supplied, we use it to expand such variable
             // declarations. This will be fixed properly in Gecko bug 1391537.
-            (&PropertyDeclaration::WithVariables(ref declaration), Some(ref computed_values)) => {
-                declaration
-                    .value
-                    .substitute_variables(
-                        declaration.id,
-                        computed_values.writing_mode,
-                        custom_properties.as_ref(),
-                        QuirksMode::NoQuirks,
-                        stylist.device(),
-                        &mut Default::default(),
-                    )
-                    .to_css(dest)
-            },
+            (&PropertyDeclaration::WithVariables(ref declaration), Some(_)) => declaration
+                .value
+                .substitute_variables(
+                    declaration.id,
+                    &context.builder.custom_properties,
+                    stylist,
+                    &context,
+                    &mut Default::default(),
+                )
+                .to_css(dest),
             (ref d, _) => d.to_css(dest),
         }
     }
@@ -904,7 +1024,7 @@ impl PropertyDeclarationBlock {
         let mut property_ids = PropertyDeclarationIdSet::default();
 
         for (property, animation_value) in animation_value_map.iter() {
-            property_ids.longhands.insert(*property);
+            property_ids.insert(property.as_borrowed());
             declarations.push(animation_value.uncompute());
         }
 
@@ -928,41 +1048,6 @@ impl PropertyDeclarationBlock {
         })
     }
 
-    /// Returns a custom properties map which is the result of cascading custom
-    /// properties in this declaration block along with context's custom
-    /// properties.
-    pub fn cascade_custom_properties_with_context(
-        &self,
-        context: &Context,
-    ) -> Option<Arc<crate::custom_properties::CustomPropertiesMap>> {
-        self.cascade_custom_properties(context.style().custom_properties(), context.style().stylist.unwrap())
-    }
-
-    /// Returns a custom properties map which is the result of cascading custom
-    /// properties in this declaration block along with the given custom
-    /// properties.
-    fn cascade_custom_properties(
-        &self,
-        inherited_custom_properties: Option<&Arc<crate::custom_properties::CustomPropertiesMap>>,
-        stylist: &Stylist,
-    ) -> Option<Arc<crate::custom_properties::CustomPropertiesMap>> {
-        let mut builder = CustomPropertiesBuilder::new(inherited_custom_properties, stylist.device());
-
-        for declaration in self.normal_declaration_iter() {
-            if let PropertyDeclaration::Custom(ref declaration) = *declaration {
-                builder.cascade(
-                    declaration,
-                    CascadePriority::new(
-                        CascadeLevel::same_tree_author_normal(),
-                        LayerOrder::root(),
-                    ),
-                );
-            }
-        }
-
-        builder.build()
-    }
-
     /// Like the method on ToCss, but without the type parameter to avoid
     /// accidentally monomorphizing this large function multiple times for
     /// different writers.
@@ -1041,7 +1126,7 @@ impl PropertyDeclarationBlock {
                 //     If all properties that map to shorthand are not present
                 //     in longhands, continue with the steps labeled shorthand
                 //     loop.
-                if !self.property_ids.longhands.contains_all(&longhands) {
+                if !self.property_ids.contains_all_longhands(&longhands) {
                     continue;
                 }
 
@@ -1387,11 +1472,18 @@ pub struct DeclarationParserState<'i> {
     importance: Importance,
     /// A list of errors that have happened so far. Not all of them might be reported.
     errors: SmallParseErrorVec<'i>,
+    /// The start of the first declaration
+    first_declaration_start: SourceLocation,
     /// The last parsed property id, if any.
     last_parsed_property_id: Option<PropertyId>,
 }
 
 impl<'i> DeclarationParserState<'i> {
+    /// Getter for first_declaration_start.
+    pub fn first_declaration_start(&self) -> SourceLocation {
+        self.first_declaration_start
+    }
+
     /// Returns whether any parsed declarations have been parsed so far.
     pub fn has_parsed_declarations(&self) -> bool {
         !self.output_block.is_empty()
@@ -1408,6 +1500,7 @@ impl<'i> DeclarationParserState<'i> {
         context: &ParserContext,
         name: CowRcStr<'i>,
         input: &mut Parser<'i, 't>,
+        declaration_start: &ParserState,
     ) -> Result<(), ParseError<'i>> {
         let id = match PropertyId::parse(&name, context) {
             Ok(id) => id,
@@ -1422,17 +1515,30 @@ impl<'i> DeclarationParserState<'i> {
             PropertyDeclaration::parse_into(&mut self.declarations, id, context, input)
         })?;
         self.importance = match input.try_parse(parse_important) {
-            Ok(()) => Importance::Important,
+            Ok(()) => {
+                if !context.allows_important_declarations() {
+                    return Err(
+                        input.new_custom_error(StyleParseErrorKind::UnexpectedImportantDeclaration)
+                    );
+                }
+                Importance::Important
+            },
             Err(_) => Importance::Normal,
         };
         // In case there is still unparsed text in the declaration, we should roll back.
         input.expect_exhausted()?;
+        let has_parsed_declarations = self.has_parsed_declarations();
         self.output_block
             .extend(self.declarations.drain(), self.importance);
         // We've successfully parsed a declaration, so forget about
         // `last_parsed_property_id`. It'd be wrong to associate any
         // following error with this property.
         self.last_parsed_property_id = None;
+
+        if !has_parsed_declarations {
+            self.first_declaration_start = declaration_start.source_location();
+        }
+
         Ok(())
     }
 
@@ -1506,8 +1612,10 @@ impl<'a, 'b, 'i> DeclarationParser<'i> for PropertyDeclarationParser<'a, 'b, 'i>
         &mut self,
         name: CowRcStr<'i>,
         input: &mut Parser<'i, 't>,
+        declaration_start: &ParserState,
     ) -> Result<(), ParseError<'i>> {
-        self.state.parse_value(self.context, name, input)
+        self.state
+            .parse_value(self.context, name, input, declaration_start)
     }
 }
 
@@ -1576,12 +1684,23 @@ fn report_one_css_error<'i>(
                 return;
             }
         }
-        error = match *property {
-            PropertyId::Custom(ref c) => {
-                StyleParseErrorKind::new_invalid(format!("--{}", c), error)
-            },
-            _ => StyleParseErrorKind::new_invalid(property.non_custom_id().unwrap().name(), error),
-        };
+        // Was able to parse property ID - Either an invalid value, or is constrained
+        // by the rule block it's in to be invalid. In the former case, we need to unwrap
+        // the error to be more specific.
+        if !matches!(
+            error.kind,
+            ParseErrorKind::Custom(StyleParseErrorKind::UnexpectedImportantDeclaration)
+        ) {
+            error = match *property {
+                PropertyId::Custom(ref c) => {
+                    StyleParseErrorKind::new_invalid(format!("--{}", c), error)
+                },
+                _ => StyleParseErrorKind::new_invalid(
+                    property.non_custom_id().unwrap().name(),
+                    error,
+                ),
+            };
+        }
     }
 
     let location = error.location;
