@@ -4,27 +4,25 @@
 
 <%namespace name="helpers" file="/helpers.mako.rs" />
 
-<% data.new_style_struct("InheritedTable", inherited=True, gecko_name="TableBorder") %>
-
 ${helpers.single_keyword(
     "border-collapse",
     "separate collapse",
-    engines="gecko servo-2013",
+    engines="gecko servo",
     gecko_enum_prefix="StyleBorderCollapse",
-    animation_value_type="discrete",
+    animation_type="discrete",
     spec="https://drafts.csswg.org/css-tables/#propdef-border-collapse",
-    servo_restyle_damage = "reflow",
+    servo_restyle_damage = "rebuild_box",
     affects="layout",
 )}
 
 ${helpers.single_keyword(
     "empty-cells",
     "show hide",
-    engines="gecko servo-2013",
+    engines="gecko servo",
     gecko_enum_prefix="StyleEmptyCells",
-    animation_value_type="discrete",
+    animation_type="discrete",
     spec="https://drafts.csswg.org/css-tables/#propdef-empty-cells",
-    servo_restyle_damage="rebuild_and_reflow",
+    servo_restyle_damage="rebuild_box",
     affects="paint",
 )}
 
@@ -32,10 +30,10 @@ ${helpers.predefined_type(
     "caption-side",
     "table::CaptionSide",
     "computed::table::CaptionSide::Top",
-    engines="gecko servo-2013",
-    animation_value_type="discrete",
+    engines="gecko servo",
+    animation_type="discrete",
     spec="https://drafts.csswg.org/css-tables/#propdef-caption-side",
-    servo_restyle_damage="rebuild_and_reflow",
+    servo_restyle_damage="rebuild_box",
     affects="layout",
 )}
 
@@ -43,11 +41,9 @@ ${helpers.predefined_type(
     "border-spacing",
     "BorderSpacing",
     "computed::BorderSpacing::zero()",
-    engines="gecko servo-2013 servo-2020",
-    servo_2020_pref="layout.2020.unimplemented",
-    animation_value_type="BorderSpacing",
+    engines="gecko servo",
     boxed=True,
     spec="https://drafts.csswg.org/css-tables/#propdef-border-spacing",
-    servo_restyle_damage="reflow",
+    servo_restyle_damage="rebuild_box",
     affects="layout",
 )}
