@@ -5,12 +5,12 @@
 //! A data structure to efficiently index structs containing selectors by local
 //! name, ids and hash.
 
-use crate::applicable_declarations::ApplicableDeclarationList;
+use crate::applicable_declarations::{ApplicableDeclarationList, ScopeProximity};
 use crate::context::QuirksMode;
 use crate::dom::TElement;
 use crate::rule_tree::CascadeLevel;
 use crate::selector_parser::SelectorImpl;
-use crate::stylist::{CascadeData, ContainerConditionId, Rule, Stylist};
+use crate::stylist::{CascadeData, ContainerConditionId, Rule, ScopeConditionId, Stylist};
 use crate::AllocErr;
 use crate::{Atom, LocalName, Namespace, ShrinkIfNeeded, WeakAtom};
 use dom::ElementState;
@@ -34,6 +34,9 @@ impl Default for PrecomputedHasher {
     }
 }
 
+/// A vector of relevant attributes, that can be useful for revalidation.
+pub type RelevantAttributes = thin_vec::ThinVec<LocalName>;
+
 /// This is a set of pseudo-classes that are both relatively-rare (they don't
 /// affect most elements by default) and likely or known to have global rules
 /// (in e.g., the UA sheets).
@@ -41,13 +44,15 @@ impl Default for PrecomputedHasher {
 /// We can avoid selector-matching those global rules for all elements without
 /// these pseudo-class states.
 const RARE_PSEUDO_CLASS_STATES: ElementState = ElementState::from_bits_retain(
-    ElementState::FULLSCREEN.bits() |
-        ElementState::VISITED_OR_UNVISITED.bits() |
-        ElementState::URLTARGET.bits() |
-        ElementState::INERT.bits() |
-        ElementState::FOCUS.bits() |
-        ElementState::FOCUSRING.bits() |
-        ElementState::TOPMOST_MODAL.bits(),
+    ElementState::FULLSCREEN.bits()
+        | ElementState::VISITED_OR_UNVISITED.bits()
+        | ElementState::URLTARGET.bits()
+        | ElementState::INERT.bits()
+        | ElementState::FOCUS.bits()
+        | ElementState::FOCUSRING.bits()
+        | ElementState::TOPMOST_MODAL.bits()
+        | ElementState::SUPPRESS_FOR_PRINT_SELECTION.bits()
+        | ElementState::HEADING_LEVEL_BITS.bits(),
 );
 
 /// A simple alias for a hashmap using PrecomputedHasher.
@@ -80,7 +85,7 @@ impl Hasher for PrecomputedHasher {
 /// A trait to abstract over a given selector map entry.
 pub trait SelectorMapEntry: Sized + Clone {
     /// Gets the selector we should use to index in the selector map.
-    fn selector(&self) -> SelectorIter<SelectorImpl>;
+    fn selector(&self) -> SelectorIter<'_, SelectorImpl>;
 }
 
 /// Map element data to selector-providing objects for which the last simple
@@ -128,8 +133,6 @@ pub struct SelectorMap<T: 'static> {
     pub rare_pseudo_classes: SmallVec<[T; 1]>,
     /// All other rules.
     pub other: SmallVec<[T; 1]>,
-    /// Whether we should bucket by attribute names.
-    bucket_attributes: bool,
     /// The number of entries in this map.
     pub count: usize,
 }
@@ -153,19 +156,10 @@ impl<T> SelectorMap<T> {
             namespace_hash: HashMap::default(),
             rare_pseudo_classes: SmallVec::new(),
             other: SmallVec::new(),
-            bucket_attributes: static_prefs::pref!("layout.css.bucket-attribute-names.enabled"),
             count: 0,
         }
     }
 
-    /// Trivially constructs an empty `SelectorMap`, with attribute bucketing
-    /// explicitly disabled.
-    pub fn new_without_attribute_bucketing() -> Self {
-        let mut ret = Self::new();
-        ret.bucket_attributes = false;
-        ret
-    }
-
     /// Shrink the capacity of the map if needed.
     pub fn shrink_if_needed(&mut self) {
         self.id_hash.shrink_if_needed();
@@ -262,21 +256,19 @@ impl SelectorMap<Rule> {
             }
         });
 
-        if self.bucket_attributes {
-            rule_hash_target.each_attr_name(|name| {
-                if let Some(rules) = self.attribute_hash.get(name) {
-                    SelectorMap::get_matching_rules(
-                        element,
-                        rules,
-                        matching_rules_list,
-                        matching_context,
-                        cascade_level,
-                        cascade_data,
-                        stylist,
-                    )
-                }
-            });
-        }
+        rule_hash_target.each_attr_name(|name| {
+            if let Some(rules) = self.attribute_hash.get(name) {
+                SelectorMap::get_matching_rules(
+                    element,
+                    rules,
+                    matching_rules_list,
+                    matching_context,
+                    cascade_level,
+                    cascade_data,
+                    stylist,
+                )
+            }
+        });
 
         if let Some(rules) = self.local_name_hash.get(rule_hash_target.local_name()) {
             SelectorMap::get_matching_rules(
@@ -340,16 +332,32 @@ impl SelectorMap<Rule> {
     ) where
         E: TElement,
     {
+        use selectors::matching::IncludeStartingStyle;
+
+        let include_starting_style = matches!(
+            matching_context.include_starting_style,
+            IncludeStartingStyle::Yes
+        );
         for rule in rules {
-            if !matches_selector(
-                &rule.selector,
-                0,
-                Some(&rule.hashes),
-                &element,
-                matching_context,
-            ) {
-                continue;
-            }
+            let scope_proximity = if rule.scope_condition_id == ScopeConditionId::none() {
+                if !matches_selector(
+                    &rule.selector,
+                    0,
+                    Some(&rule.hashes),
+                    &element,
+                    matching_context,
+                ) {
+                    continue;
+                }
+                ScopeProximity::infinity()
+            } else {
+                let result =
+                    cascade_data.find_scope_proximity_if_matching(rule, element, matching_context);
+                if result == ScopeProximity::infinity() {
+                    continue;
+                }
+                result
+            };
 
             if rule.container_condition_id != ContainerConditionId::none() {
                 if !cascade_data.container_condition_matches(
@@ -362,7 +370,22 @@ impl SelectorMap<Rule> {
                 }
             }
 
-            matching_rules.push(rule.to_applicable_declaration_block(cascade_level, cascade_data));
+            if rule.is_starting_style {
+                // Set this flag if there are any rules inside @starting-style. This flag is for
+                // optimization to avoid any redundant resolution of starting style if the author
+                // doesn't specify for this element.
+                matching_context.has_starting_style = true;
+
+                if !include_starting_style {
+                    continue;
+                }
+            }
+
+            matching_rules.push(rule.to_applicable_declaration_block(
+                cascade_level,
+                cascade_data,
+                scope_proximity,
+            ));
         }
     }
 }
@@ -388,8 +411,8 @@ impl<T: SelectorMapEntry> SelectorMap<T> {
                         .class_hash
                         .try_entry(class.clone(), quirks_mode)?
                         .or_default(),
-                    Bucket::Attribute { name, lower_name } |
-                    Bucket::LocalName { name, lower_name } => {
+                    Bucket::Attribute { name, lower_name }
+                    | Bucket::LocalName { name, lower_name } => {
                         // If the local name in the selector isn't lowercase,
                         // insert it into the rule hash twice. This means that,
                         // during lookup, we can always find the rules based on
@@ -432,11 +455,7 @@ impl<T: SelectorMapEntry> SelectorMap<T> {
 
         let bucket = {
             let mut disjoint_buckets = SmallVec::new();
-            let bucket = find_bucket(
-                entry.selector(),
-                &mut disjoint_buckets,
-                self.bucket_attributes,
-            );
+            let bucket = find_bucket(entry.selector(), &mut disjoint_buckets);
 
             // See if inserting this selector in multiple entries in the
             // selector map would be worth it. Consider a case like:
@@ -453,8 +472,8 @@ impl<T: SelectorMapEntry> SelectorMap<T> {
             // This is specially true if there's any universal selector in the
             // `disjoint_selectors` set, at which point we'd just be doing
             // wasted work.
-            if !disjoint_buckets.is_empty() &&
-                disjoint_buckets
+            if !disjoint_buckets.is_empty()
+                && disjoint_buckets
                     .iter()
                     .all(|b| b.more_specific_than(&bucket))
             {
@@ -481,12 +500,24 @@ impl<T: SelectorMapEntry> SelectorMap<T> {
     ///
     /// FIXME(bholley) This overlaps with SelectorMap<Rule>::get_all_matching_rules,
     /// but that function is extremely hot and I'd rather not rearrange it.
-    pub fn lookup<'a, E, F>(&'a self, element: E, quirks_mode: QuirksMode, f: F) -> bool
+    pub fn lookup<'a, E, F>(
+        &'a self,
+        element: E,
+        quirks_mode: QuirksMode,
+        relevant_attributes: Option<&mut RelevantAttributes>,
+        f: F,
+    ) -> bool
     where
         E: TElement,
         F: FnMut(&'a T) -> bool,
     {
-        self.lookup_with_state(element, element.state(), quirks_mode, f)
+        self.lookup_with_state(
+            element,
+            element.state(),
+            quirks_mode,
+            relevant_attributes,
+            f,
+        )
     }
 
     #[inline]
@@ -495,6 +526,7 @@ impl<T: SelectorMapEntry> SelectorMap<T> {
         element: E,
         element_state: ElementState,
         quirks_mode: QuirksMode,
+        mut relevant_attributes: Option<&mut RelevantAttributes>,
         mut f: F,
     ) -> bool
     where
@@ -538,24 +570,25 @@ impl<T: SelectorMapEntry> SelectorMap<T> {
             return false;
         }
 
-        if self.bucket_attributes {
-            element.each_attr_name(|name| {
-                if done {
-                    return;
+        element.each_attr_name(|name| {
+            if done {
+                return;
+            }
+            if let Some(v) = self.attribute_hash.get(name) {
+                if let Some(ref mut relevant_attributes) = relevant_attributes {
+                    relevant_attributes.push(name.clone());
                 }
-                if let Some(v) = self.attribute_hash.get(name) {
-                    for entry in v.iter() {
-                        if !f(&entry) {
-                            done = true;
-                            return;
-                        }
+                for entry in v.iter() {
+                    if !f(&entry) {
+                        done = true;
+                        return;
                     }
                 }
-            });
-
-            if done {
-                return false;
             }
+        });
+
+        if done {
+            return false;
         }
 
         if let Some(v) = self.local_name_hash.get(element.local_name()) {
@@ -617,6 +650,7 @@ impl<T: SelectorMapEntry> SelectorMap<T> {
             element,
             element.state() | additional_states,
             quirks_mode,
+            /* relevant_attributes = */ None,
             |entry| f(entry),
         ) {
             return false;
@@ -697,26 +731,23 @@ type DisjointBuckets<'a> = SmallVec<[Bucket<'a>; 5]>;
 fn specific_bucket_for<'a>(
     component: &'a Component<SelectorImpl>,
     disjoint_buckets: &mut DisjointBuckets<'a>,
-    bucket_attributes: bool,
 ) -> Bucket<'a> {
     match *component {
         Component::Root => Bucket::Root,
         Component::ID(ref id) => Bucket::ID(id),
         Component::Class(ref class) => Bucket::Class(class),
-        Component::AttributeInNoNamespace { ref local_name, .. } if bucket_attributes => {
-            Bucket::Attribute {
-                name: local_name,
-                lower_name: local_name,
-            }
+        Component::AttributeInNoNamespace { ref local_name, .. } => Bucket::Attribute {
+            name: local_name,
+            lower_name: local_name,
         },
         Component::AttributeInNoNamespaceExists {
             ref local_name,
             ref local_name_lower,
-        } if bucket_attributes => Bucket::Attribute {
+        } => Bucket::Attribute {
             name: local_name,
             lower_name: local_name_lower,
         },
-        Component::AttributeOther(ref selector) if bucket_attributes => Bucket::Attribute {
+        Component::AttributeOther(ref selector) => Bucket::Attribute {
             name: &selector.local_name,
             lower_name: &selector.local_name_lower,
         },
@@ -745,18 +776,14 @@ fn specific_bucket_for<'a>(
         //
         // So inserting `span` in the rule hash makes sense since we want to
         // match the slotted <span>.
-        Component::Slotted(ref selector) => {
-            find_bucket(selector.iter(), disjoint_buckets, bucket_attributes)
-        },
-        Component::Host(Some(ref selector)) => {
-            find_bucket(selector.iter(), disjoint_buckets, bucket_attributes)
-        },
+        Component::Slotted(ref selector) => find_bucket(selector.iter(), disjoint_buckets),
+        Component::Host(Some(ref selector)) => find_bucket(selector.iter(), disjoint_buckets),
         Component::Is(ref list) | Component::Where(ref list) => {
             if list.len() == 1 {
-                find_bucket(list.slice()[0].iter(), disjoint_buckets, bucket_attributes)
+                find_bucket(list.slice()[0].iter(), disjoint_buckets)
             } else {
                 for selector in list.slice() {
-                    let bucket = find_bucket(selector.iter(), disjoint_buckets, bucket_attributes);
+                    let bucket = find_bucket(selector.iter(), disjoint_buckets);
                     disjoint_buckets.push(bucket);
                 }
                 Bucket::Universal
@@ -782,13 +809,12 @@ fn specific_bucket_for<'a>(
 fn find_bucket<'a>(
     mut iter: SelectorIter<'a, SelectorImpl>,
     disjoint_buckets: &mut DisjointBuckets<'a>,
-    bucket_attributes: bool,
 ) -> Bucket<'a> {
     let mut current_bucket = Bucket::Universal;
 
     loop {
         for ss in &mut iter {
-            let new_bucket = specific_bucket_for(ss, disjoint_buckets, bucket_attributes);
+            let new_bucket = specific_bucket_for(ss, disjoint_buckets);
             // NOTE: When presented with the choice of multiple specific selectors, use the
             // rightmost, on the assumption that that's less common, see bug 1829540.
             if new_bucket.more_or_equally_specific_than(&current_bucket) {
@@ -833,7 +859,7 @@ impl<V> MaybeCaseInsensitiveHashMap<Atom, V> {
         &mut self,
         mut key: Atom,
         quirks_mode: QuirksMode,
-    ) -> Result<hash_map::Entry<Atom, V>, AllocErr> {
+    ) -> Result<hash_map::Entry<'_, Atom, V>, AllocErr> {
         if quirks_mode == QuirksMode::Quirks {
             key = key.to_ascii_lowercase()
         }
@@ -848,7 +874,7 @@ impl<V> MaybeCaseInsensitiveHashMap<Atom, V> {
     }
 
     /// HashMap::iter
-    pub fn iter(&self) -> hash_map::Iter<Atom, V> {
+    pub fn iter(&self) -> hash_map::Iter<'_, Atom, V> {
         self.0.iter()
     }
 
