@@ -17,9 +17,10 @@ use crate::selector_parser::{PseudoElement, SelectorImpl};
 use crate::stylist::RuleInclusion;
 use log::Level::Trace;
 use selectors::matching::{
-    MatchingForInvalidation, MatchingContext, MatchingMode, NeedsSelectorFlags,
-    RelativeSelectorMatchingState, VisitedHandlingMode,
+    IncludeStartingStyle, MatchingContext, MatchingForInvalidation, MatchingMode,
+    NeedsSelectorFlags, VisitedHandlingMode,
 };
+use selectors::parser::PseudoElement as PseudoElementTrait;
 use servo_arc::Arc;
 
 /// Whether pseudo-elements should be resolved or not.
@@ -48,11 +49,20 @@ where
 struct MatchingResults {
     rule_node: StrongRuleNode,
     flags: ComputedValueFlags,
+    has_starting_style: bool,
 }
 
 /// A style returned from the resolver machinery.
 pub struct ResolvedStyle(pub Arc<ComputedValues>);
 
+impl ResolvedStyle {
+    /// Convenience accessor for the style.
+    #[inline]
+    pub fn style(&self) -> &ComputedValues {
+        &*self.0
+    }
+}
+
 /// The primary style of an element or an element-backed pseudo-element.
 pub struct PrimaryStyle {
     /// The style itself.
@@ -60,6 +70,11 @@ pub struct PrimaryStyle {
     /// Whether the style was reused from another element via the rule node (see
     /// `StyleSharingCache::lookup_by_rules`).
     pub reused_via_rule_node: bool,
+    /// The element may have matched rules inside @starting-style.
+    /// Basically, we don't apply @starting-style rules to |style|. This is a sugar to let us know
+    /// if we should resolve the element again for starting style, which is the after-change style
+    /// with @starting-style rules applied in addition.
+    pub may_have_starting_style: bool,
 }
 
 /// A set of style returned from the resolver machinery.
@@ -80,6 +95,12 @@ impl ResolvedElementStyles {
     pub fn primary_style_mut(&mut self) -> &mut Arc<ComputedValues> {
         &mut self.primary.style.0
     }
+
+    /// Returns true if this element may have starting style rules.
+    #[inline]
+    pub fn may_have_starting_style(&self) -> bool {
+        self.primary.may_have_starting_style
+    }
 }
 
 impl PrimaryStyle {
@@ -98,7 +119,7 @@ impl From<ResolvedElementStyles> for ElementStyles {
     }
 }
 
-fn with_default_parent_styles<E, F, R>(element: E, f: F) -> R
+pub(crate) fn with_default_parent_styles<E, F, R>(element: E, f: F) -> R
 where
     E: TElement,
     F: FnOnce(Option<&ComputedValues>, Option<&ComputedValues>) -> R,
@@ -143,16 +164,16 @@ fn eager_pseudo_is_definitely_not_generated(
 
     if !style
         .flags
-        .intersects(ComputedValueFlags::DISPLAY_DEPENDS_ON_INHERITED_STYLE) &&
-        style.get_box().clone_display() == Display::None
+        .intersects(ComputedValueFlags::DISPLAY_DEPENDS_ON_INHERITED_STYLE)
+        && style.get_box().clone_display() == Display::None
     {
         return true;
     }
 
     if !style
         .flags
-        .intersects(ComputedValueFlags::CONTENT_DEPENDS_ON_INHERITED_STYLE) &&
-        style.ineffective_content_property()
+        .intersects(ComputedValueFlags::CONTENT_DEPENDS_ON_INHERITED_STYLE)
+        && style.ineffective_content_property()
     {
         return true;
     }
@@ -187,16 +208,22 @@ where
         &mut self,
         parent_style: Option<&ComputedValues>,
         layout_parent_style: Option<&ComputedValues>,
+        include_starting_style: IncludeStartingStyle,
     ) -> PrimaryStyle {
-        let primary_results = self.match_primary(VisitedHandlingMode::AllLinksUnvisited);
+        let primary_results = self.match_primary(
+            VisitedHandlingMode::AllLinksUnvisited,
+            include_starting_style,
+        );
 
         let inside_link = parent_style.map_or(false, |s| s.visited_style().is_some());
 
-        let visited_rules = if self.context.shared.visited_styles_enabled &&
-            (inside_link || self.element.is_link())
+        let visited_rules = if self.context.shared.visited_styles_enabled
+            && (inside_link || self.element.is_link())
         {
-            let visited_matching_results =
-                self.match_primary(VisitedHandlingMode::RelevantLinkVisited);
+            let visited_matching_results = self.match_primary(
+                VisitedHandlingMode::RelevantLinkVisited,
+                IncludeStartingStyle::No,
+            );
             Some(visited_matching_results.rule_node)
         } else {
             None
@@ -210,6 +237,8 @@ where
             },
             parent_style,
             layout_parent_style,
+            include_starting_style,
+            primary_results.has_starting_style,
         )
     }
 
@@ -218,12 +247,15 @@ where
         inputs: CascadeInputs,
         parent_style: Option<&ComputedValues>,
         layout_parent_style: Option<&ComputedValues>,
+        include_starting_style: IncludeStartingStyle,
+        may_have_starting_style: bool,
     ) -> PrimaryStyle {
         // Before doing the cascade, check the sharing cache and see if we can
         // reuse the style via rule node identity.
-        let may_reuse = self.element.matches_user_and_content_rules() &&
-            parent_style.is_some() &&
-            inputs.rules.is_some();
+        let may_reuse = self.element.matches_user_and_content_rules()
+            && parent_style.is_some()
+            && inputs.rules.is_some()
+            && include_starting_style == IncludeStartingStyle::No;
 
         if may_reuse {
             let cached = self.context.thread_local.sharing_cache.lookup_by_rules(
@@ -250,6 +282,7 @@ where
                 /* pseudo = */ None,
             ),
             reused_via_rule_node: false,
+            may_have_starting_style,
         }
     }
 
@@ -259,11 +292,16 @@ where
         parent_style: Option<&ComputedValues>,
         layout_parent_style: Option<&ComputedValues>,
     ) -> ResolvedElementStyles {
-        let primary_style = self.resolve_primary_style(parent_style, layout_parent_style);
+        let primary_style =
+            self.resolve_primary_style(parent_style, layout_parent_style, IncludeStartingStyle::No);
 
         let mut pseudo_styles = EagerPseudoStyles::default();
 
-        if !self.element.is_pseudo_element() {
+        if !self
+            .element
+            .implemented_pseudo_element()
+            .is_some_and(|p| !PseudoElementTrait::is_element_backed(&p))
+        {
             let layout_parent_style_for_pseudo =
                 layout_parent_style_for_pseudo(&primary_style, layout_parent_style);
             SelectorImpl::each_eagerly_cascaded_pseudo_element(|pseudo| {
@@ -274,8 +312,8 @@ where
                 );
 
                 if let Some(style) = pseudo_style {
-                    if !matches!(self.pseudo_resolution, PseudoElementResolution::Force) &&
-                        eager_pseudo_is_definitely_not_generated(&pseudo, &style.0)
+                    if !matches!(self.pseudo_resolution, PseudoElementResolution::Force)
+                        && eager_pseudo_is_definitely_not_generated(&pseudo, &style.0)
                     {
                         return;
                     }
@@ -342,19 +380,16 @@ where
     ) -> ResolvedStyle {
         debug_assert!(pseudo.map_or(true, |p| p.is_eager()));
 
-        let implemented_pseudo = self.element.implemented_pseudo_element();
-        let pseudo = pseudo.or(implemented_pseudo.as_ref());
-
         let mut conditions = Default::default();
         let values = self.context.shared.stylist.cascade_style_and_visited(
             Some(self.element),
             pseudo,
             inputs,
             &self.context.shared.guards,
-            pseudo.and(parent_style),
             parent_style,
             layout_parent_style,
             FirstLineReparenting::No,
+            /* try_tactic = */ &Default::default(),
             Some(&self.context.thread_local.rule_cache),
             &mut conditions,
         );
@@ -373,10 +408,16 @@ where
     pub fn cascade_styles_with_default_parents(
         &mut self,
         inputs: ElementCascadeInputs,
+        may_have_starting_style: bool,
     ) -> ResolvedElementStyles {
         with_default_parent_styles(self.element, move |parent_style, layout_parent_style| {
-            let primary_style =
-                self.cascade_primary_style(inputs.primary, parent_style, layout_parent_style);
+            let primary_style = self.cascade_primary_style(
+                inputs.primary,
+                parent_style,
+                layout_parent_style,
+                IncludeStartingStyle::No,
+                may_have_starting_style,
+            );
 
             let mut pseudo_styles = EagerPseudoStyles::default();
             if let Some(mut pseudo_array) = inputs.pseudos.into_array() {
@@ -398,8 +439,8 @@ where
                             Some(&pseudo),
                         );
 
-                        if !matches!(self.pseudo_resolution, PseudoElementResolution::Force) &&
-                            eager_pseudo_is_definitely_not_generated(&pseudo, &style.0)
+                        if !matches!(self.pseudo_resolution, PseudoElementResolution::Force)
+                            && eager_pseudo_is_definitely_not_generated(&pseudo, &style.0)
                         {
                             continue;
                         }
@@ -425,6 +466,7 @@ where
         let MatchingResults {
             rule_node,
             mut flags,
+            has_starting_style: _,
         } = self.match_pseudo(
             &originating_element_style.style.0,
             pseudo,
@@ -457,7 +499,11 @@ where
         ))
     }
 
-    fn match_primary(&mut self, visited_handling: VisitedHandlingMode) -> MatchingResults {
+    fn match_primary(
+        &mut self,
+        visited_handling: VisitedHandlingMode,
+        include_starting_style: IncludeStartingStyle,
+    ) -> MatchingResults {
         debug!(
             "Match primary for {:?}, visited: {:?}",
             self.element, visited_handling
@@ -471,17 +517,17 @@ where
             Some(bloom_filter),
             selector_caches,
             visited_handling,
+            include_starting_style,
             self.context.shared.quirks_mode(),
             NeedsSelectorFlags::Yes,
             MatchingForInvalidation::No,
         );
 
         let stylist = &self.context.shared.stylist;
-        let implemented_pseudo = self.element.implemented_pseudo_element();
         // Compute the primary rule node.
         stylist.push_applicable_declarations(
             self.element,
-            implemented_pseudo.as_ref(),
+            None,
             self.element.style_attribute(),
             self.element.smil_override(),
             self.element.animation_declarations(self.context.shared),
@@ -506,28 +552,11 @@ where
                 }
             }
         }
-        // This is a bit awkward - ideally, the flag is set directly where `considered_relative_selector`
-        // is; however, in that context, the implementation detail of `extra_data` is not visible, so
-        // it's done here. A trait for manipulating the flags is an option, but not worth it for a single flag.
-        match matching_context.considered_relative_selector {
-            RelativeSelectorMatchingState::None => (),
-            RelativeSelectorMatchingState::Considered => {
-                matching_context
-                    .extra_data
-                    .cascade_input_flags
-                    .insert(ComputedValueFlags::CONSIDERED_RELATIVE_SELECTOR);
-            },
-            RelativeSelectorMatchingState::ConsideredAnchor => {
-                matching_context.extra_data.cascade_input_flags.insert(
-                    ComputedValueFlags::ANCHORS_RELATIVE_SELECTOR |
-                        ComputedValueFlags::CONSIDERED_RELATIVE_SELECTOR,
-                );
-            },
-        };
 
         MatchingResults {
             rule_node,
             flags: matching_context.extra_data.cascade_input_flags,
+            has_starting_style: matching_context.has_starting_style,
         }
     }
 
@@ -542,10 +571,6 @@ where
             self.element, pseudo_element, visited_handling
         );
         debug_assert!(pseudo_element.is_eager());
-        debug_assert!(
-            !self.element.is_pseudo_element(),
-            "Element pseudos can't have any other eager pseudo."
-        );
 
         let mut applicable_declarations = ApplicableDeclarationList::new();
 
@@ -566,6 +591,7 @@ where
             Some(bloom_filter),
             selector_caches,
             visited_handling,
+            IncludeStartingStyle::No,
             self.context.shared.quirks_mode(),
             NeedsSelectorFlags::Yes,
             MatchingForInvalidation::No,
@@ -596,6 +622,75 @@ where
         Some(MatchingResults {
             rule_node,
             flags: matching_context.extra_data.cascade_input_flags,
+            has_starting_style: false, // We don't care.
         })
     }
+
+    /// Resolve the starting style.
+    pub fn resolve_starting_style(&mut self) -> PrimaryStyle {
+        // Compute after-change style for the parent and the layout parent.
+        // Per spec, starting style inherits from the parent’s after-change style just like
+        // after-change style does.
+        let parent_el = self.element.inheritance_parent();
+        let parent_data = parent_el.as_ref().and_then(|e| e.borrow_data());
+        let parent_style = parent_data.as_ref().map(|d| d.styles.primary());
+        let parent_after_change_style = parent_style.and_then(|s| self.after_change_style(s));
+        let parent_values = parent_after_change_style
+            .as_ref()
+            .or(parent_style)
+            .map(|x| &**x);
+
+        let mut layout_parent_el = parent_el.clone();
+        let layout_parent_data;
+        let layout_parent_after_change_style;
+        let layout_parent_values = if parent_style.map_or(false, |s| s.is_display_contents()) {
+            layout_parent_el = Some(layout_parent_el.unwrap().layout_parent());
+            layout_parent_data = layout_parent_el.as_ref().unwrap().borrow_data().unwrap();
+            let layout_parent_style = Some(layout_parent_data.styles.primary());
+            layout_parent_after_change_style =
+                layout_parent_style.and_then(|s| self.after_change_style(s));
+            layout_parent_after_change_style
+                .as_ref()
+                .or(layout_parent_style)
+                .map(|x| &**x)
+        } else {
+            parent_values
+        };
+
+        self.resolve_primary_style(
+            parent_values,
+            layout_parent_values,
+            IncludeStartingStyle::Yes,
+        )
+    }
+
+    /// If there is no transition rule in the ComputedValues, it returns None.
+    pub fn after_change_style(
+        &mut self,
+        primary_style: &Arc<ComputedValues>,
+    ) -> Option<Arc<ComputedValues>> {
+        let rule_node = primary_style.rules();
+        let without_transition_rules = self
+            .context
+            .shared
+            .stylist
+            .rule_tree()
+            .remove_transition_rule_if_applicable(rule_node);
+        if without_transition_rules == *rule_node {
+            // We don't have transition rule in this case, so return None to let
+            // the caller use the original ComputedValues.
+            return None;
+        }
+
+        // FIXME(bug 868975): We probably need to transition visited style as
+        // well.
+        let inputs = CascadeInputs {
+            rules: Some(without_transition_rules),
+            visited_rules: primary_style.visited_rules().cloned(),
+            flags: primary_style.flags.for_cascade_inputs(),
+        };
+
+        let style = self.cascade_style_and_visited_with_default_parents(inputs);
+        Some(style.0)
+    }
 }
